from typing import List, Dict, Union, Optional, TypeVar, Generic

import pytest

from pytractions.base import (
    Base,
    JSONIncompatibleError,
    TList,
    TDict,
    JSON_COMPATIBLE,
    TypeNode,
    NoAnnotationError,
)

# Jsonable test cases

T = TypeVar("T")
T2 = TypeVar("T")


def test_base_jsonable_basic_ok():
    class TestC(Base):
        i: int
        s: str


def test_base_jsonable_basic_union_ok():
    class TestC(Base):
        i: Union[int, str]


def test_base_jsonable_basic_optional():
    class TestC(Base):
        i: Optional[int]


def test_base_jsonable_basic_structured_ok():
    class TestC(Base):
        l: TList[int]
        d: TDict[str, int]


def test_base_jsonable_basic_structured_optional_ok():
    class TestC(Base):
        l: Optional[TList[int]]
        d: TDict[str, Optional[int]]


def test_base_jsonable_basic_structured_union_ok():
    class TestC(Base):
        l: TList[int]
        d: TDict[str, Union[int, str]]


def test_base_jsonable_basic_nested_structured_ok():
    class TestC(Base):
        l: TList[TDict[str, int]]


def test_base_jsonable_basic_nested_structured_optional_ok():
    class TestC(Base):
        l: TList[Optional[TDict[str, Optional[int]]]]


def test_base_jsonable_basic_nested_structured_union_ok():
    class TestC(Base):
        l: TList[Union[TDict[str, int], TDict[str, str]]]


def test_base_jsonable_basic_clasess_ok():
    class TestC1(Base):
        i: int

    class TestC(Base):
        c: TestC1


def test_base_jsonable_basic_clasess_optional_ok():
    class TestC1(Base):
        i: int

    class TestC(Base):
        c: Optional[TestC1]


def test_base_jsonable_basic_clasess_union_ok():
    class TestC1(Base):
        i: int

    class TestC(Base):
        c: TDict[str, Union[TestC1, int]]


def test_generic_jsonable_ok():
    class TestC1(Base, Generic[T]):
        x: T


def test_generic_jsonable_concrete_ok():
    class TestC1(Base, Generic[T]):
        x: T

    TestC1[int]


# Jsonable expected to fail cases


def test_base_jsonable_basic_fail():
    with pytest.raises(JSONIncompatibleError):

        class TestC(Base):
            i: tuple
            s: str


def test_base_jsonable_basic_union_fail():
    with pytest.raises(JSONIncompatibleError):

        class TestC(Base):
            i: Union[int, tuple]


def test_base_jsonable_basic_structured_fail():
    with pytest.raises(JSONIncompatibleError):

        class TestC(Base):
            l: List[int]
            d: Dict[str, tuple]


def test_base_jsonable_basic_structured_union_fail():
    with pytest.raises(JSONIncompatibleError):

        class TestC(Base):
            l: List[int]
            d: Dict[str, Union[tuple, str]]


def test_base_jsonable_basic_nested_structured_fail():
    with pytest.raises(JSONIncompatibleError):

        class TestC(Base):
            l: List[Dict[str, tuple]]


def test_base_jsonable_basic_nested_structured_union_fail():
    with pytest.raises(JSONIncompatibleError):

        class TestC(Base):
            l: List[Union[Dict[str, tuple], Dict[str, str]]]


def test_base_jsonable_basic_clasess_fail():
    class TestC1:
        i: int

    with pytest.raises(JSONIncompatibleError):

        class TestC(Base):
            c: TestC1


def test_base_jsonable_basic_clasess_union_fail():
    class TestC1:
        i: int

    with pytest.raises(JSONIncompatibleError):

        class TestC(Base):
            c: Dict[str, Union[TestC1, int]]


def test_generic_jsonable_concrete_fail():
    class X:
        pass

    class TestC1(Base, Generic[T]):
        x: T

    with pytest.raises(JSONIncompatibleError):
        TestC1[X]


# setattr validation test cases


def test_base_setattr_ok():
    class TestC(Base):
        i: int
        s: str

    t = TestC(i=10, s="a")
    t.i = 100
    t.s = "a"


def test_base_setattr_optional_ok():
    class TestC(Base):
        s: Optional[str]

    t = TestC(s="a")
    t.s = None


def test_base_setattr_union_ok():
    class TestC(Base):
        s: Union[str, int]

    t = TestC(s="a")
    t.s = 10


def test_base_setattr_complex_ok():
    class TestC(Base):
        s: TList[int]

    t = TestC(s=TList[int]([5]))
    t.s = TList[int]([10])


def test_base_setattr_complex2_ok():
    class TestC(Base, Generic[T]):
        td: TDict[str, T]

    d: TDict[str, int] = TDict[str, int]({"a": 10})

    t = TestC[int](td=TDict[str, int]({"b": 30}))
    t.td = d


# setattr validation test cases - fail


def test_base_setattr_fail():
    class TestC(Base):
        i: int
        s: str

    t = TestC(i=10, s="a")
    with pytest.raises(TypeError):
        t.i = "a"
    with pytest.raises(TypeError):
        t.s = 100


def test_base_setattr_missing_annotation_fail():
    with pytest.raises(NoAnnotationError):

        class TestC(Base):
            i = 10


def test_base_setattr_no_attr_fail():
    class TestC(Base):
        i: int
        s: str

    t = TestC(i=10, s="a")
    with pytest.raises(AttributeError):
        t.a = "a"


def test_base_setattr_optional_fail():
    class TestC(Base):
        s: Optional[str]

    t = TestC(s="a")
    print("---")
    t.s = None
    print("---")
    with pytest.raises(TypeError):
        t.s = 100


def test_base_setattr_union_fail():
    class TestC(Base):
        s: Union[str, int]

    t = TestC(s=10)
    with pytest.raises(TypeError):
        t.s = {}


def test_base_setattr_complex_fail():
    class TestC(Base):
        s: TList[str]

    t = TestC(s=TList[str](["a"]))
    with pytest.raises(TypeError):
        t.s = [10]


def test_base_setattr_complex1_fail():
    class TestC(Base):
        s: TList[int]

    l: TList[str] = TList[str](["a"])
    t = TestC(s=TList[int]([10]))
    with pytest.raises(TypeError):
        t.s = l


def test_base_setattr_complex2_fail():
    class TestList(Base, Generic[T]):
        l: TList[T]

    class TestC(Base):
        tl: TestList[int]

    _list = TList[str](["a"])

    tl2: TestList[str] = TestList[str](l=_list)

    t = TestC(tl=TestList[int](l=TList[int]([20])))
    with pytest.raises(TypeError):
        t.tl = tl2


def test_base_setattr_complex3_ok():

    class TestDict(Base, Generic[T]):
        d: TDict[str, T]

    class TestC(Base):
        td: TestDict[int]

    d: Dict[str, int] = {"a": 10}

    with pytest.raises(TypeError):
        TestDict[int](d=d)

    d2: TDict[str, int] = TDict[str, int]({"a": 10})
    d3: TDict[str, str] = TDict[str, str]({"a": "a"})

    td: TestDict[int] = TestDict[int](d=d2)
    t = TestC(td=td)
    td2: TestDict[str] = TestDict[str](d=d3)
    with pytest.raises(TypeError):
        t.td = td2


def test_base_generic_nested():
    class TestA(Base):
        pass

    class TestC(Base, Generic[T]):
        a: T

    TestC[TestA](a=TestA())


# Type testing


def test_type_json():
    t = TypeNode.from_type(TList[JSON_COMPATIBLE])
    tjson = t.to_json()
    print(tjson)
    t2 = TypeNode.from_json(tjson)
    assert t == t2


def test_base_type_to_json():
    class TestC(Base):
        i: int
        s: str

    assert TestC.type_to_json() == {
        "$type": {
            "args": [],
            "module": "tests.test_base",
            "type": "test_base_type_to_json.<locals>.TestC",
        },
        "i": {
            "$type": {
                "args": [],
                "module": "builtins",
                "type": "int",
            },
            "default": None,
        },
        "s": {
            "$type": {
                "args": [],
                "module": "builtins",
                "type": "str",
            },
            "default": None,
        },
    }


def test_base_type_nested_to_json():
    class TestA(Base):
        a: str

    class TestC(Base):
        i: int
        s: str
        ta: TestA

    assert TestC.type_to_json() == {
        "$type": {
            "args": [],
            "module": "tests.test_base",
            "type": "test_base_type_nested_to_json.<locals>.TestC",
        },
        "i": {
            "$type": {
                "args": [],
                "module": "builtins",
                "type": "int",
            },
            "default": None,
        },
        "s": {
            "$type": {
                "args": [],
                "module": "builtins",
                "type": "str",
            },
            "default": None,
        },
        "ta": {
            "$type": {
                "args": [],
                "module": "tests.test_base",
                "type": "test_base_type_nested_to_json.<locals>.TestA",
            },
            "a": {
                "$type": {
                    "args": [],
                    "module": "builtins",
                    "type": "str",
                },
                "default": None,
            },
        },
    }


# Test Generics


def test_generic_wrong_param_count():
    class TestC(Base, Generic[T, T2]):
        a: T
        b: T2

    with pytest.raises(TypeError):
        TestC[int]



import streamlit as st

st.title("Pytractions fundamentals")
st.markdown(
    """
Fundamentals
------------

Pytractions is python frameworks aimed to simplify the processs of creating complex data pipelines
based on reusable basic blocks connected with each other by IOs. Single block of the pipeline is
called Traction.

Traction constist of 4 public type of arguments: `inputs`, `outputs`, `resources` and `args`

Traction input can be connected to other traction output. Like following (pseudocode):
```pseudocode
Model:
   name: str
   age: int

T1:
    output1: Out[Model]

T2:
    input1: In[Model]

T2.input1 = T1.output1
```
One input can be connected to one output. However one output can be connected to multiple inputs.
Input and output classes are containers for atomic data.
Input attribute has to be connected only to  output (or input) attribute of the same type.
It's not possible to connect input to inner structure defined within an output.
As defined above, it's not possible to do (pseudocode):
```
T3
  input: In[str]

T3.input = T1.output1.name
```

### Core ideas

* Traction constist of 4 public type of arguments: `inputs`, `outputs`, `resources` and `args`
* Traction consists of public attribute definitions and definion of `run` method which processes
inputs and produces outputs. Anything else (if needed) should be declared as private.
* Traction `run` method is supposed to process input data and produce output data.
Obtaining, storing or updating data to external resources should be done using `resources`.
* All used external functionality is defined in `resources` and passed to traction when it's
initialized to make testing easier.
* Traction can have zero or more inputs and zero or more outputs. It's prefered to have more
outputs than
one complex output. More outputs makes traction easier to connect to other tractions.
As example, traction FoodSorter is better defined with outputs: `meat`, `vegetables`, `fruits`
rather than with output `sorted_food: {meat: [], vegetables: [], fruits: []}`

* After traction process is finished, all information about it's execution are available in
traction instance attributes.
* Everything publically available within a traction is json compatible
  [typing-system](#typing-system).

Typing system
-------------
Tractions are supposed to work with non-binary data which are consisting of json compatible
elementary types. These types are: `bool`, `string`, `int`, `float`, `list` and `dict`.
Supported is also any
combination of these types and any json compatible class consisted of these types.
More information about typing is available here [typing](developer_guide#pytraction-base).

Traction attributes
-------------------

**Inputs and outputs**

Inputs and ouputs are supposed to be json compatible types of models consistent of these which
holds information about the data which are supposed to be processed by traction. Input or output
shouldn't be a class providing methods for data manipulation - those should be rather handled
by traction itself or by a resource.
As an example of wrong input data consider `MyHttpClient` which holds information about
http session configuration and provide also functionality to fetch http pages.
Session should be rather defined as resource which should be

1. Fully initialized as a resource; inputs should be requested urls. Pseudo code below:
```
mhc = MyHttpClient(auth=(...), retries=(...), timeout=(...))

traction run method:
    for input_url in inputs:
        mhc.fetch(input_url)
```

2. Partially initiated as a resource and final configuration should be done per traction
run via traction arguments; inputs should be requested urls.
```
mhc = MyHttpClient(auth=(...))

traction run method:
    retries = args.retries
    timeout = args.timeout
    for input_url in inputs:
        mhc.fetch(input_url, retries=retries, timeout=timeout)
```

3. Partially initialized and final configuration should be passed as input model which
consists of request attributes and url;
```
mhc = MyHttpClient()

traction run method:
    retries = args.retries
    timeout = args.timeout
    for request_data in inputs:
        mhc.fetch(request.data_url,
                  auth=request_data.auth,
                  retries=request_data.retries,
                  timeout=request_data.timeout)
```

Arguments
---------
Any information which is is constant for whole traction process should be defined as argument.
Those can be for example: number of retries, timeout, threshold, number of threads etc

Resources
---------
Resources ment to provide access to external funcionality to generate, update or store data.
Resource class should be written to support elementary operations for the data.
Any complex operations on the
data should be handled by the traction itself.
Example use cases for a resource can be:
- Fetch client data form a database
- Create-Or-Update client data in database
- Delete client data from database

Other use cases should be defined in tractions.
As example of traction we can define: `RemoveInactiveClients`
which should do following:
- Fetch all clients from database
- Filter out inactive clients
- Filter clients scheduled for removal
- For all clients scheduled for ramoval
    - Remove client data from database (via the client resource)
- For all remaning inactive clients
    - Send notifications to inactive clients (via a resource)
    - Set client data to scheduled for removal
    - Udate client data in database (via the client resource)

Joining tractions together
--------------------------
Tration can be connected to each other by t2.o_input1 = t1.i_input.
But to makes things reusable there's
special tration class called `Tractor`. Idea of tractor is simple: define all tractions and their
connections in one class which can be later used as regular traction. Schema is following:
```
Tractor
    i_input: In[int] = In[int](data=0)
    a_arg: Arg[int] = Arg[int](data=0)
    t_traction1: Traction1 = Traction1(uid='1', i_input=i_input, a_arg=a_arg)
    t_traction2: Traction2 = Traction2(uid='2', i_input=t_traction1.o_output, a_arg=a_arg)
    o_output: Out[int] = t_traction2.o_output
```
When tractor is initialized, all tractions are copied from class variables into instance variables
located in `Tractor.traction` list.
You can learn details about tractors in [tractor](developer_guide#tractors) section.
"""
)



import pytest

from pytractions.base import (
    Base,
    Traction,
    Arg,
    In,
    Out,
    Res,
    OnUpdateCallable,
)
from pytractions.exc import TractionFailedError


class NOOPResource(Base):
    """NOOP Resource."""

    pass


def test_tractor_attr():

    # wrong doc attribute
    with pytest.raises(TypeError):

        class TestTraction1(Traction):
            i_input: In[int]
            d_: int

    # wrong doc attr attribute
    with pytest.raises(TypeError):

        class TestTraction2(Traction):
            i_input: In[int]
            d_i_input: int

    # custom attribute
    with pytest.raises(TypeError):

        class TestTraction3(Traction):
            custom_attribute: int


def test_to_json_from_json():
    class TestTraction(Traction):
        i_input: In[int]
        o_output: Out[int]
        r_res: Res[NOOPResource]
        a_arg: Arg[str]

        def _run(self, on_update: OnUpdateCallable) -> None:
            self.o_output.data = self.i_input.data

    t = TestTraction(
        uid="test-traction-1",
        i_input=In[int](data=1),
        a_arg=Arg[str](a="test"),
        r_res=Res[NOOPResource](r=NOOPResource()),
    )
    assert t.to_json() == {
        "$data": {
            "a_arg": {
                "$data": {
                    "a": "test",
                },
                "$type": {
                    "args": [
                        {
                            "args": [],
                            "module": "builtins",
                            "type": "str",
                        },
                    ],
                    "module": "pytractions.base",
                    "type": "Arg",
                },
            },
            "details": {
                "$data": {},
                "$type": {
                    "args": [
                        {
                            "args": [],
                            "module": "builtins",
                            "type": "str",
                        },
                        {
                            "args": [],
                            "module": "builtins",
                            "type": "str",
                        },
                    ],
                    "module": "pytractions.base",
                    "type": "TDict",
                },
            },
            "errors": {
                "$data": [],
                "$type": {
                    "args": [
                        {
                            "args": [],
                            "module": "builtins",
                            "type": "str",
                        },
                    ],
                    "module": "pytractions.base",
                    "type": "TList",
                },
            },
            "i_input": {
                "$data": {
                    "data": 1,
                },
                "$type": {
                    "args": [
                        {
                            "args": [],
                            "module": "builtins",
                            "type": "int",
                        },
                    ],
                    "module": "pytractions.base",
                    "type": "In",
                },
            },
            "o_output": {
                "$data": {
                    "data": 0,
                },
                "$type": {
                    "args": [
                        {
                            "args": [],
                            "module": "builtins",
                            "type": "int",
                        },
                    ],
                    "module": "pytractions.base",
                    "type": "Out",
                },
            },
            "r_res": {
                "$data": {
                    "r": {
                        "$data": {},
                        "$type": {
                            "args": [],
                            "module": "tests.test_traction",
                            "type": "NOOPResource",
                        },
                    },
                },
                "$type": {
                    "args": [
                        {
                            "args": [],
                            "module": "tests.test_traction",
                            "type": "NOOPResource",
                        },
                    ],
                    "module": "pytractions.base",
                    "type": "Res",
                },
            },
            "skip": False,
            "skip_reason": "",
            "state": "ready",
            "stats": {
                "$data": {
                    "finished": "",
                    "skipped": False,
                    "started": "",
                },
                "$type": {
                    "args": [],
                    "module": "pytractions.base",
                    "type": "TractionStats",
                },
            },
            "uid": "test-traction-1",
        },
        "$type": {
            "args": [],
            "module": "tests.test_traction",
            "type": "test_to_json_from_json.<locals>.TestTraction",
        },
    }
    t2 = TestTraction.from_json(t.to_json(), _locals=locals())
    assert t == t2


def test_to_run_failed():
    class TestTraction(Traction):
        i_input: In[int]
        o_output: Out[int]
        r_res: Res[NOOPResource]
        a_arg: Arg[str]

        def _run(self, on_update: OnUpdateCallable) -> None:
            self.o_output.data = self.i_input.data
            raise TractionFailedError

    t = TestTraction(
        uid="test-traction-1",
        i_input=In[int](data=1),
        a_arg=Arg[str](a="test"),
        r_res=Res[NOOPResource](r=NOOPResource()),
    )
    t.run()
    assert t.state == "failed"


def test_to_run_error():
    class TestTraction(Traction):
        i_input: In[int]
        o_output: Out[int]
        r_res: Res[NOOPResource]
        a_arg: Arg[str]

        def _run(self, on_update: OnUpdateCallable) -> None:
            self.o_output.data = self.i_input.data
            raise ValueError("test error")

    t = TestTraction(
        uid="test-traction-1",
        i_input=In[int](data=1),
        a_arg=Arg[str](a="test"),
        r_res=Res[NOOPResource](r=NOOPResource()),
    )

    with pytest.raises(ValueError):
        t.run()
    assert t.state == "error"



import argparse
import re

import importlib.metadata


def filter_name(name_filter, name):
    """Test if name matches filter_name."""
    if isinstance(name_filter, re.Pattern):
        return re.match(name_filter, name)
    return name_filter in name


def filter_tag(tag_filter, tags):
    """Test if tag is in list of tags."""
    if isinstance(tag_filter, re.Pattern):
        return any(re.match(tag_filter, t) for t in tags)
    return tag_filter in tags


ignore_types = (
    "TList",
    "TDict",
    "Union",
    "Optional",
    "In",
    "Res",
    "Out",
    "TRes",
    "Arg",
    "STMDSingleIn",
    "ABase",
    "TIn",
)


def gather_simple_types(type_json):
    """Gather simple types from complex type in json format."""
    simple_types = []
    if type_json["type"]["$type"]["type"] not in ignore_types:
        simple_types.append(type_json["type"]["$type"]["type"])
    stack = []
    for arg in type_json["type"]["$type"].get("args", []):
        stack.append(arg)
    while stack:
        arg = stack.pop()
        if arg["type"] not in ignore_types:
            simple_types.append(arg["type"])
        for a in arg.get("args", []):
            stack.append(a)
    return list(set(simple_types))


def tractions_discovery(
    tag_filter=None,
    type_filter=None,
    name_filter=None,
    inputs_type_filter=None,
    outputs_type_filter=None,
    resources_type_filter=None,
    args_type_filter=None,
    type_to_str=False,
):
    """Inspect all available distributions for tractions entry points."""
    ret = []
    seen = set()
    for d in importlib.metadata.distributions():
        if d.entry_points.select(group="tractions"):
            if d._path in seen:
                continue
            seen.add(d._path)
            dist = inspect_distribution(
                d,
                tag_filter=tag_filter,
                type_filter=type_filter,
                name_filter=name_filter,
                inputs_type_filter=inputs_type_filter,
                outputs_type_filter=outputs_type_filter,
                resources_type_filter=resources_type_filter,
                args_type_filter=args_type_filter,
                type_to_str=type_to_str,
            )
            if dist and dist not in ret:
                ret.append(dist)
    return ret


def _type_to_str(type_json):
    """Return string representation for given type."""
    out = []
    stack = []
    out.append(type_json["$type"]["type"])
    if type_json["$type"].get("args", []):
        stack.insert(0, "]")
        for n, arg in enumerate(type_json["$type"].get("args", [])[::-1]):
            stack.insert(0, arg)
            if n != len(type_json["$type"].get("args", [])) - 1:
                stack.insert(0, ",")
        stack.insert(0, "[")

    while stack:
        item = stack.pop(0)
        if isinstance(item, str):
            out.append(item)
            continue
        out.append(item["type"])
        if item.get("args", []):
            stack.insert(0, "]")
            for n, a in enumerate(item.get("args", [])[::-1]):
                stack.insert(0, a)
                if n != len(item.get("args", [])) - 1:
                    stack.insert(0, ",")
            stack.insert(0, "[")
    return "".join(out)


def convert_types_to_str(t_meta):
    """Return string representation of attribute type."""
    t_meta["inputs"] = [
        {"name": i["name"], "type": _type_to_str(i["type"]), "docs": str(i["docs"])}
        for i in t_meta["inputs"]
    ]
    t_meta["outputs"] = [
        {"name": i["name"], "type": _type_to_str(i["type"]), "docs": str(i["docs"])}
        for i in t_meta["outputs"]
    ]
    t_meta["resources"] = [
        {"name": i["name"], "type": _type_to_str(i["type"]), "docs": str(i["docs"])}
        for i in t_meta["resources"]
    ]
    t_meta["args"] = [
        {"name": i["name"], "type": _type_to_str(i["type"]), "docs": str(i["docs"])}
        for i in t_meta["args"]
    ]
    return t_meta


def inspect_traction_ep(traction_ep, tag_filter=None, name_filter=None, type_filter=None):
    """Inspect traction entry point."""
    t = {}
    traction = traction_ep.load()
    t["name"] = str(traction)
    t["type"] = traction._TYPE
    t["module"] = traction.__module__
    t["docs"] = getattr(traction, "d_", None)
    t["tags"] = []

    if name_filter and not filter_name(name_filter, t["name"]):
        return {}
    if tag_filter and not filter_tag(tag_filter, traction.tags):
        return {}
    if type_filter and traction._TYPE not in type_filter:
        return {}

    t["inputs"] = [
        {"name": k, "type": v.type_to_json(), "docs": getattr(traction, "d_" + k, None)}
        for k, v in traction._fields.items()
        if k.startswith("i_")
    ]
    t["outputs"] = [
        {"name": k, "type": v.type_to_json(), "docs": getattr(traction, "d_" + k, None)}
        for k, v in traction._fields.items()
        if k.startswith("o_")
    ]
    t["resources"] = [
        {"name": k, "type": v.type_to_json(), "docs": getattr(traction, "d_" + k, None)}
        for k, v in traction._fields.items()
        if k.startswith("r_")
    ]
    t["args"] = [
        {"name": k, "type": v.type_to_json(), "docs": getattr(traction, "d_" + k, None)}
        for k, v in traction._fields.items()
        if k.startswith("a_")
    ]
    return t


def inspect_distribution(
    distribution,
    tag_filter=None,
    type_filter=None,
    name_filter=None,
    inputs_type_filter=None,
    outputs_type_filter=None,
    resources_type_filter=None,
    args_type_filter=None,
    type_to_str=False,
):
    """Inspect python distribution object for tractions entry points."""
    d = {}
    d["name"] = distribution.metadata["Name"]
    d["tags"] = []
    d["tractions"] = []
    d["args"] = []
    d["inputs"] = []
    d["outputs"] = []
    d["resources"] = []
    for t in distribution.entry_points.select(group="tractions"):
        t_meta = inspect_traction_ep(
            t, tag_filter=tag_filter, type_filter=type_filter, name_filter=name_filter
        )
        if t_meta:
            d["args"].extend(sum([gather_simple_types(a) for a in t_meta["args"]], []))
            d["resources"].extend(sum([gather_simple_types(a) for a in t_meta["resources"]], []))
            d["inputs"].extend(sum([gather_simple_types(a) for a in t_meta["inputs"]], []))
            d["outputs"].extend(sum([gather_simple_types(a) for a in t_meta["outputs"]], []))
            d["args"] = list(set(d["args"]))
            d["resources"] = list(set(d["resources"]))
            d["inputs"] = list(set(d["inputs"]))
            d["outputs"] = list(set(d["outputs"]))
        else:
            continue

        t_meta_str_types = convert_types_to_str(t_meta.copy())

        if inputs_type_filter:
            if not any(
                [
                    any([rt in t["type"] for t in t_meta_str_types["inputs"]])
                    for rt in inputs_type_filter
                ]
            ):
                continue
        if outputs_type_filter:
            if not any(
                [
                    any([rt in t["type"] for t in t_meta_str_types["outputs"]])
                    for rt in outputs_type_filter
                ]
            ):
                continue
        if resources_type_filter:
            if not any(
                [
                    any([rt in t["type"] for t in t_meta_str_types["resources"]])
                    for rt in resources_type_filter
                ]
            ):
                continue
        if args_type_filter:
            if not any(
                [
                    any([rt in t["type"] for t in t_meta_str_types["args"]])
                    for rt in args_type_filter
                ]
            ):
                continue
        d["tractions"].append(t_meta_str_types if type_to_str else t_meta)

    return d


def catalog(
    tag_filter=None,
    type_filter=None,
    name_filter=None,
    args_type_filter=None,
    inputs_type_filter=None,
    outputs_type_filter=None,
    resources_type_filter=None,
    type_to_str=False,
):
    """Return list of available tractions based on given filters."""
    tractions = tractions_discovery(
        tag_filter=tag_filter,
        type_filter=type_filter,
        name_filter=name_filter,
        args_type_filter=args_type_filter,
        inputs_type_filter=inputs_type_filter,
        outputs_type_filter=outputs_type_filter,
        resources_type_filter=resources_type_filter,
        type_to_str=type_to_str,
    )
    all_inputs = list(set([inp for t in tractions for inp in t["inputs"]]))
    all_outputs = list(set([out for t in tractions for out in t["outputs"]]))
    all_resources = list(set([res for t in tractions for res in t["resources"]]))
    all_args = list(set([arg for t in tractions for arg in t["args"]]))
    tags = []
    return tractions, all_inputs, all_outputs, all_resources, all_args, tags


def catalog_main(args):
    """Run catalog for CLI."""
    tractions = catalog(
        tag_filter=args.tag_filter,
        type_filter=args.type_filter,
        name_filter=args.name_filter,
        type_to_str=True,
    )
    if args.output == "json":
        import json

        print(json.dumps(tractions[0], indent=2))
    elif args.output == "yaml":
        import yaml

        print(yaml.dump(tractions[0]))


def make_parsers(subparsers):
    """Make parser for catalog subcommand."""
    p_catalog = subparsers.add_parser("catalog", help="Explore all locally available tractions")
    p_catalog.add_argument(
        "--tag", "-t", dest="tag_filter", help="Filter tractions by tag", default=None
    )
    p_catalog.add_argument(
        "--type",
        "-y",
        dest="type_filter",
        help="Filter tractions by type",
        choices=["TRACTOR", "TRACTION", "STMD"],
        default=None,
    )
    p_catalog.add_argument(
        "--name", "-n", dest="name_filter", help="Filter by distribution name", default=None
    )
    p_catalog.add_argument(
        "--output-format",
        "-o",
        dest="output",
        help="Output format",
        choices=["json", "yaml"],
        default="json",
    )
    p_catalog.set_defaults(command=catalog_main)
    return p_catalog


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Pytraction container ep")
    subparsers = parser.add_subparsers(required=True, dest="command")
    make_parsers(subparsers)
    args = parser.parse_args()
    tractions, _, _, _, _, _ = catalog(
        tag_filter=args.tag_filter,
        type_filter=args.type_filter,
        name_filter=args.name_filter,
        type_to_str=True,
    )
    if args.output == "json":
        import json

        print(json.dumps(tractions, indent=2))
    elif args.output == "yaml":
        import yaml

        print(yaml.dump(tractions))



import os
from streamlit.web import cli as stcli
import sys

dirname = os.path.dirname(__file__)
filename = os.path.join(dirname, "welcome.py")


def main(args):
    """Run server for web interface."""
    sys.exit(stcli.main_run([filename], standalone_mode=False))


def make_parsers(subparsers):
    """Parsers for this module."""
    p_web = subparsers.add_parser("web", help="Run web interface")
    p_web.set_defaults(command=main)
    return p_web



from __future__ import annotations
import abc
import dataclasses

from typing import get_origin, get_args


class ANYMeta(abc.ABCMeta):
    """Metaclass for helper object that compares equal to everything."""

    def __eq__(mcs, other):
        """Return always True."""
        return True

    def __ne__(mcs, other):
        """Return always False."""
        return False

    def __repr__(mcs):
        """Any class string representation."""
        return "<ANY>"

    def __hash__(mcs):
        """Return id of the class."""
        return id(mcs)


class ANY(metaclass=ANYMeta):
    """A helper object that compares equal to everything."""

    def __eq__(cls, other):
        """Return True."""
        return True

    def __ne__(cls, other):
        """Return always False."""
        return False

    def __repr__(cls):
        """Return Any class string representation."""
        return "<ANY>"

    def __hash__(cls):
        """Return id of the class."""
        return id(cls)


def doc(docstring: str):
    """Create dataclass field for doctring fields."""
    return dataclasses.field(init=False, repr=False, default=docstring)


# class OType:
#
#     origins: Any
#     args: List["OType" | Any]
#     parameters: List["OType" | Any]
#
#     def __init__(self):
#         self.origins = []
#         self.args = []
#         self.parameters = []
#
#     def __str__(self):
#         return f"OType({self.origins}, {self.args}, {self.parameters})"
#
#     def __eq__(self, other):
#         print("OType.__eq__")
#         return (
#             self.origins == other.origins
#             and self.args == other.args
#             and self.parameters == other.parameters
#         )
#
#     def __le__(self, other):
#         if self.origins in ([list], [List]) and other.origins in ([list], [List[ANY]]):
#             # TODO: fix list check
#             olte = True
#             return True
#         else:
#             olte = self.origins == other.origins or any(
#                 [issubclass(o1, o2) for o1, o2 in zip(self.origins, other.origins)]
#             )
#         alte = True
#         if len(self.args) != len(other.args):
#             alte = False
#         else:
#             for a1, a2 in zip(self.args, other.args):
#                 if isinstance(a1, OType) and isinstance(a2, OType):
#                     alte &= a1 <= a2
#                 elif not isinstance(a1, OType) and not isinstance(a2, OType):
#                     alte &= any([issubclass(o1, o2) for o1, o2 in
#                                  zip(self.origins, other.origins)])
#                 else:
#                     alte = False
#                     break
#         return olte and alte
#
#     def __check_generics__(self, other):
#         olte = self.origins == other.origins or any(
#             [issubclass(o1, o2) for o1, o2 in zip(self.origins, other.origins)]
#         )
#         alte = True
#         if len(self.args) != len(other.parameters):
#             alte = False
#         else:
#             alte = True
#         return olte and alte


def _get_args(v):
    return get_args(v) or v.__targs__ if hasattr(v, "__targs__") else []


def _get_origin(v):
    return get_origin(v) or v.__torigin__ if hasattr(v, "__torigin__") else None


# def get_type(var):
#     root = OType()
#     if _get_origin(var) == Union:
#         root.origins = _get_origin(get_args(var))
#     elif _get_origin(var):
#         root.origins = [_get_origin(var)]
#     else:
#         root.origins = [var]
#
#     root.parameters = getattr(var, "__parameters__", [])
#
#     to_process = []
#     if _get_args(var):
#         for arg in _get_args(var):
#             to_process.append((root, arg))
#
#     while to_process:
#         croot, arg = to_process.pop(0)
#         child = OType()
#         child.parameters = arg.__parameters__ if hasattr(arg, "__parameters__") else []
#         if _get_origin(arg) == Union:
#             child.origins = _get_args(_get_origin(arg))
#         elif _get_origin(arg):
#             child.origins = [_get_origin(arg)]
#         else:
#             child.origins = [arg]
#
#         if _get_args(arg):
#             for charg in get_args(arg):
#                 to_process.append((child, charg))
#         croot.args.append(child)
#
#     return root
#
#
# def check_type(to_check, expected):
#     t1 = get_type(to_check)
#     t2 = get_type(expected)
#     return t1 <= t2
#
#
# def check_type_generics(to_check, expected_generic):
#     return get_type(to_check).__check_generics__(get_type(expected_generic))



from typing import TypeVar, Generic, Union

import pytest

from pytractions.base import Base, TList, TDict

# Jsonable test cases

T = TypeVar("T")


def test_base_list_new_ok():
    l: TList[int] = TList[int]([10])
    assert l._list == [10]


def test_base_list_new_ok_complex():
    l: TList[TDict[int, str]] = TList[TDict[int, str]]([TDict[int, str]({10: "a"})])
    assert l._list == [TDict[int, str]({10: "a"})]


def test_base_list_new_ok_generic():
    class TestC(Base, Generic[T]):
        l: TList[T]

    tc = TestC[int](l=TList[int]([10]))
    assert tc.l._list == [10]


def test_base_list_append_ok():
    l: TList[int] = TList[int]([10])
    l.append(20)

    assert l._list == [10, 20]


def test_base_list_extend_ok():
    l: TList[int] = TList[int]([10])
    l.extend(TList[int]([30, 40]))

    assert l._list == [10, 30, 40]


def test_base_list_add_ok():
    l: TList[int] = TList[int]([10])
    l2 = l + TList[int]([30, 40])
    assert l2._list == [10, 30, 40]


def test_base_list_insert_ok():
    l: TList[int] = TList[int]([10])
    l.insert(0, 5)

    assert l._list == [5, 10]


def test_base_list_setitem_ok():
    l: TList[int] = TList[int]([10])
    l[0] = 5

    assert l._list == [5]


def test_contains_ok():
    l: TList[int] = TList[int]([10])
    assert 10 in l


def test_base_list_new_fail():
    with pytest.raises(TypeError):
        TList[int](["a"])


def test_base_list_new_complex_fail():
    with pytest.raises(TypeError):
        TList[TDict[int, str]]([TDict[int, int]({10: 10})])


def test_base_list_new_generic_fail():
    class TestC(Base, Generic[T]):
        l: TList[T]

    with pytest.raises(TypeError):
        TestC[int](l=TList[int](["a"]))


def test_base_list_new_generic_no_type_fail():
    with pytest.raises(TypeError):
        TList([1, 2, 3])


def test_base_list_append_fail():
    l: TList[int] = TList[int]([10])
    with pytest.raises(TypeError):
        l.append("a")


def test_base_list_extend_fail():
    l: TList[int] = TList[int]([10])
    with pytest.raises(TypeError):
        l.extend(TList[str](["a", "a"]))


def test_base_list_insert_fail():
    l: TList[int] = TList[int]([10])
    with pytest.raises(TypeError):
        l.insert(0, "a")


def test_base_list_setitem_fail():
    l: TList[int] = TList[int]([10])
    with pytest.raises(TypeError):
        l[0] = "a"


def test_contains_fail():
    l: TList[int] = TList[int]([10])
    with pytest.raises(TypeError):
        assert "a" in l


def test_base_list_add_fail():
    l: TList[int] = TList[int]([10])
    with pytest.raises(TypeError):
        l + TList[str](["a", "a"])


def test_base_list_len():
    l: TList[int] = TList[int]([10, 20, 30])

    assert len(l) == 3


def test_del_item():
    l: TList[int] = TList[int]([10, 20, 30])
    del l[0]
    assert l._list == [20, 30]


def test_iter():
    l: TList[int] = TList[int]([10, 20, 30])
    ret = []
    for x in l:
        ret.append(x)
    assert ret == [10, 20, 30]


def test_clear():
    l: TList[int] = TList[int]([10, 20, 30])
    l.clear()
    assert l._list == []


def test_count():
    l: TList[int] = TList[int]([10, 20, 30])
    assert l.count(10) == 1


def test_index():
    l: TList[int] = TList[int]([10, 20, 30])
    assert l.index(20) == 1


class TestC(Base):
    """Test class for complex test cases."""

    x: int
    y: TDict[str, Union[str, TList[str]]]


def test_index_complex():
    l: TList[TestC] = TList[TestC](
        [
            TestC(
                x=10,
                y=TDict[str, Union[str, TList[str]]](
                    {"test": "test1", "foo": TList[str](["a", "b"])}
                ),
            ),
            TestC(
                x=20,
                y=TDict[str, Union[str, TList[str]]](
                    {"test": "test2", "foo": TList[str](["c", "d"])}
                ),
            ),
            TestC(
                x=30,
                y=TDict[str, Union[str, TList[str]]](
                    {"test": "test3", "foo": TList[str](["e", "f"])}
                ),
            ),
        ]
    )
    print(l)
    assert (
        l.index(
            TestC(
                x=20,
                y=TDict[str, Union[str, TList[str]]](
                    {"test": "test2", "foo": TList[str](["c", "d"])}
                ),
            )
        )
        == 1
    )


def test_insert():
    l: TList[int] = TList[int]([10, 20, 30])
    l.insert(2, 25)
    assert l._list == [10, 20, 25, 30]


def test_pop():
    l: TList[int] = TList[int]([10, 20, 25, 30])
    assert l.pop(2) == 25
    assert l._list == [10, 20, 30]


def test_remove():
    l: TList[int] = TList[int]([10, 20, 25, 30])
    l.remove(25)
    assert l._list == [10, 20, 30]


def test_reverse():
    l: TList[int] = TList[int]([10, 20, 30])
    l.reverse()
    assert l._list == [30, 20, 10]


def test_sort():
    l: TList[int] = TList[int]([10, 1, 20, 5, 30])
    l.sort()
    assert l._list == [1, 5, 10, 20, 30]



import inspect
import importlib
import json
from typing import get_origin, TypeVar, Union, ForwardRef, Dict, Any, Tuple, List
import sys


from .abase import ABase
from .utils import ANY

X = TypeVar("X")

JSON_COMPATIBLE = Union[int, float, str, bool, ABase, type(None), X]


def evaluate_forward_ref(ref, frame):
    """Evaluate ForwardRef."""
    caller_globals, caller_locals = frame.f_globals, frame.f_locals
    recursive_guard = set()
    return ref._evaluate(caller_globals, caller_locals, recursive_guard)


class CMPNode:
    """CMPNode class for TypeNode comparison."""

    def __init__(self, n1, n2, op, ch_op):
        """Initialize CMPNode."""
        self.n1 = n1
        self.n2 = n2
        self.children = []
        self.op = op
        self.eq = None
        self.ch_op = ch_op

    def __str__(self):
        """Return string representation of CMPNode."""
        return "<CMPNode n1=%s n2=%s op=%s eq=%s>" % (self.n1, self.n2, self.op, self.eq)

    def __repr__(self):
        """Return repr representation of CMPNode."""
        return "<CMPNode n1=%s n2=%s op=%s eq=%s>" % (self.n1, self.n2, self.op, self.eq)


class TypeNode:
    """TypeNode class for type comparison."""

    _json_cache = []
    _to_type_cache = {}

    def __str__(self):
        """Return string representation of TypeNode."""
        return "<TypeNode type=%s>" % self.type_

    def __init__(self, type_, subclass_check=True):
        """Initialize TypeNode."""
        self.type_ = type_
        self.subclass_check = subclass_check
        self.children = []

    @classmethod
    def from_type(cls, type_, subclass_check=True):
        """Create TypeNode from provided type)."""
        root = cls(type_=type_, subclass_check=subclass_check)
        current = root
        stack = []
        while True:
            if hasattr(current.type_, "__args__"):
                for arg in current.type_.__args__:
                    n = cls(type_=arg)
                    stack.append(n)
                    current.children.append(n)
            elif hasattr(current.type_, "_params"):
                for arg in current.type_._params:
                    n = cls(type_=arg)
                    stack.append(n)
                    current.children.append(n)
            if not stack:
                break
            if get_origin(current.type_):
                current.type_ = get_origin(current.type_)
            elif hasattr(current.type_, "_orig_cls") and current.type_._orig_cls:
                current.type_ = current.type_._orig_cls
            current = stack.pop()
        return root

    # def post_order(self):
    #     stack = [(self, 0, None)]
    #     post_order = []
    #     while stack:
    #         current, parent_index, current_parent = stack.pop(0)
    #         for n, ch in enumerate(current.children):
    #             stack.insert(0, (ch, n, current))
    #         post_order.insert(0, (current, parent_index, current_parent))
    #     return post_order

    def replace_params(self, params_map):
        """Replace Typevars in TypeNode structure with values from provided mapping."""
        stack = [(self, 0, None)]
        while stack:
            current, parent_index, current_parent = stack.pop(0)
            for n, ch in enumerate(current.children):
                stack.insert(0, (ch, n, current))
            if type(current.type_) is TypeVar:
                if current.type_ in params_map:
                    if hasattr(params_map[current.type_], "__args__"):
                        replaced_tn = TypeNode.from_type(params_map[current.type_])
                        current.type_ = replaced_tn.type_
                        current.children = replaced_tn.children

                    else:
                        current.type_ = params_map[current.type_]

    def to_type(self, types_cache={}, params_map={}):
        """Return new type for TypeNode or already existing type from cache."""
        if json.dumps(self.to_json(), sort_keys=True) in self._to_type_cache:
            return self._to_type_cache[json.dumps(self.to_json(), sort_keys=True)]

        stack = [(self, 0, None)]
        post_order = []
        while stack:
            current, parent_index, current_parent = stack.pop(0)
            for n, ch in enumerate(current.children):
                stack.insert(0, (ch, n, current))
            post_order.insert(0, (current, parent_index, current_parent))

        for item in post_order:
            node, parent_index, parent = item

            if node.children:
                type_ = node.type_

                # children_types = tuple([x.type_ for x in node.children])
                children_type_ids = tuple([id(x.type_) for x in node.children])

                if f"{type_.__qualname__}[{children_type_ids}]" in types_cache:
                    node.type_ = types_cache[f"{type_.__qualname__}[{children_type_ids}]"]
                else:
                    if type_ == Union:
                        node.type_ = Union[tuple([x.type_ for x in node.children])]
                    else:
                        new_type = type_.__class_getitem__(
                            tuple([x.type_ for x in node.children]), params_map=params_map
                        )
                        node.type_ = new_type

            if not parent:
                continue
            parent.children[parent_index] = node
        self._to_type_cache[json.dumps(self.to_json(), sort_keys=True)] = post_order[-1][0].type_
        return post_order[-1][0].type_

    @staticmethod
    def __determine_op(ch1, ch2) -> str:
        op = "all"
        if (ch1.type_ is Union and ch2.type_ is Union) or (
            ch1.type_ is not Union and ch2.type_ is not Union
        ):
            op = "all"
        elif (ch1.type_ is not Union and ch2.type_ is Union) or (
            ch1.type_ is Union and ch2.type_ is not Union
        ):
            op = "any"
        return op

    def __eq_post_order(self, root_node):
        stack = [root_node]
        post_order = []
        post_order.insert(0, root_node)
        while stack:
            current_node = stack.pop()

            if current_node.n1.type_ is Union and current_node.n2.type_ is not Union:
                for ch1 in current_node.n1.children:
                    node = CMPNode(ch1, current_node.n2, "all", "all")
                    stack.insert(0, node)
                    post_order.insert(0, node)
                    current_node.children.append(node)

            elif current_node.n1.type_ is not Union and current_node.n2.type_ is Union:
                for ch2 in current_node.n2.children:
                    node = CMPNode(current_node.n1, ch2, "all", "all")
                    stack.insert(0, node)
                    post_order.insert(0, node)
                    current_node.children.append(node)

            elif current_node.op == "all":
                for ch1, ch2 in zip(current_node.n1.children, current_node.n2.children):
                    op = self.__determine_op(ch1, ch2)
                    node = CMPNode(ch1, ch2, op, op)
                    stack.insert(0, node)
                    post_order.insert(0, node)
                    current_node.children.append(node)
            else:
                if current_node.n1.type_ is Union:
                    for ch in current_node.n1.children:
                        op = self.__determine_op(ch, current_node.n2.type_)
                        node = CMPNode(ch, current_node.n2, op, op)
                        stack.insert(0, node)
                        post_order.insert(0, node)
                        current_node.children.append(node)
                else:
                    for ch in current_node.n2.children:
                        op = self.__determine_op(ch, current_node.n1.type_)
                        node = CMPNode(ch, current_node.n1, op, op)
                        stack.insert(0, node)
                        post_order.insert(0, node)
                        current_node.children.append(node)
        return post_order

    def __eq__(self, other):
        """Test if TypeNode is Equal to other TypeNode."""
        if type(other) is not TypeNode:
            return False

        op = self.__determine_op(self, other)
        node = CMPNode(self, other, op, op)
        post_order = self.__eq_post_order(node)

        for cmp_node in post_order:
            if cmp_node.op == "any":
                if cmp_node.children:
                    ch_eq = any([ch.eq for ch in cmp_node.children])
                else:
                    ch_eq = True
            else:
                ch_eq = all([ch.eq for ch in cmp_node.children])

            n1_type = get_origin(cmp_node.n1.type_) or cmp_node.n1.type_
            n2_type = get_origin(cmp_node.n2.type_) or cmp_node.n2.type_

            if isinstance(n1_type, ForwardRef):
                frame = sys._getframe(1)
                n1_type = evaluate_forward_ref(n1_type, frame)
            if isinstance(n2_type, ForwardRef):
                frame = sys._getframe(1)
                n2_type = evaluate_forward_ref(n2_type, frame)
            if n2_type is None:
                n2_type = type(None)
            if n1_type is None:
                n1_type = type(None)

            # check types only of both types are not union
            # otherwise equality was already decided by check above

            orig_cls1 = getattr(n1_type, "_orig_cls", False)
            orig_cls2 = getattr(n2_type, "_orig_cls", True)

            has_params1 = hasattr(n1_type, "_params") and len(n1_type._params) != 0
            has_params2 = hasattr(n2_type, "_params") and len(n2_type._params) != 0

            if n1_type != Union and n2_type != Union:
                ch_eq &= (
                    n1_type == n2_type
                    or (orig_cls1 == orig_cls2 and has_params1 and has_params2)
                    or orig_cls1 == n2_type
                    or orig_cls2 == n1_type
                    or (
                        self.subclass_check
                        and (
                            inspect.isclass(n1_type)
                            and inspect.isclass(n2_type)
                            and issubclass(n1_type, n2_type)
                        )
                    )
                    or (
                        self.subclass_check
                        and (
                            inspect.isclass(n1_type)
                            and inspect.isclass(orig_cls2)
                            and issubclass(n1_type, orig_cls2)
                        )
                    )
                    or bool(
                        self.subclass_check
                        and inspect.isclass(orig_cls1)
                        and inspect.isclass(orig_cls2)
                        and issubclass(orig_cls1, orig_cls2)
                    )
                )
            cmp_node.eq = ch_eq

        return node.eq

    def __eq_cached_(self, other):
        if type(other) is not TypeNode:
            return False

        op = self.__determine_op(self, other)
        node = CMPNode(self, other, op, op)
        post_order = self.__eq_post_order(node)

        for cmp_node in post_order:
            if cmp_node.op == "any":
                if cmp_node.children:
                    ch_eq = any([ch.eq for ch in cmp_node.children])
                else:
                    ch_eq = True
            else:
                ch_eq = all([ch.eq for ch in cmp_node.children])

            n1_type = cmp_node.n1.type_
            n2_type = cmp_node.n2.type_

            if isinstance(n1_type, ForwardRef) or isinstance(n2_type, ForwardRef):
                frame = sys._getframe(1)

            if isinstance(n1_type, ForwardRef):
                n1_type = evaluate_forward_ref(n1_type, frame)
            if isinstance(n2_type, ForwardRef):
                n2_type = evaluate_forward_ref(n2_type, frame)

            if n2_type is None:
                n2_type = type(None)
            if n1_type is None:
                n1_type = type(None)

            # check types only of both types are not union
            # otherwise equality was already decided by check above

            orig_cls1 = getattr(n1_type, "_orig_cls", False)
            orig_cls2 = getattr(n2_type, "_orig_cls", True)

            has_params1 = hasattr(n1_type, "_params") and len(n1_type._params) != 0
            has_params2 = hasattr(n2_type, "_params") and len(n2_type._params) != 0

            is_any1 = n1_type is ANY
            is_any2 = n2_type is ANY

            if is_any1 or is_any2:
                ch_eq &= is_any1 and is_any2
            elif n1_type is not Union and n2_type is not Union:
                ch_eq &= (
                    n1_type == n2_type
                    or (orig_cls1 == orig_cls2 and has_params1 and has_params2)
                    or orig_cls1 == n2_type
                    or orig_cls2 == n1_type
                    or (
                        self.subclass_check
                        and (
                            inspect.isclass(n1_type)
                            and inspect.isclass(n2_type)
                            and issubclass(n1_type, n2_type)
                        )
                    )
                    or (
                        self.subclass_check
                        and (
                            inspect.isclass(n1_type)
                            and inspect.isclass(orig_cls2)
                            and issubclass(n1_type, orig_cls2)
                        )
                    )
                    or bool(
                        self.subclass_check
                        and inspect.isclass(orig_cls1)
                        and inspect.isclass(orig_cls2)
                        and issubclass(orig_cls1, orig_cls2)
                    )
                )
            else:
                ch_eq = False
            cmp_node.eq = ch_eq

        return node.eq

    def json_compatible(self):
        """Check if type in TypeNode is json compatible."""
        if self.children:
            op = "all"
        else:
            op = "any"

        root_node = CMPNode(self, TypeNode(JSON_COMPATIBLE), op, op)
        stack = [root_node]
        post_order = []
        post_order.insert(0, root_node)
        while stack:
            current_node = stack.pop()
            if current_node.n1.children:
                for ch1 in current_node.n1.children:
                    if ch1.children:
                        op = "all"
                    else:
                        op = "any"
                    node = CMPNode(ch1, TypeNode(JSON_COMPATIBLE), op, op)
                    stack.insert(0, node)
                    post_order.insert(0, node)
                    current_node.children.append(node)
            else:
                for u in (int, float, str, bool, ABase, type(None), TypeVar("X")):
                    node = CMPNode(current_node.n1, TypeNode(u), "any", "any")
                    post_order.insert(0, node)
                    current_node.children.append(node)

        for cmp_node in post_order:
            n1_type = get_origin(cmp_node.n1.type_) or cmp_node.n1.type_
            n2_type = get_origin(cmp_node.n2.type_) or cmp_node.n2.type_
            if isinstance(n1_type, ForwardRef):
                frame = sys._getframe(1)
                n1_type = evaluate_forward_ref(n1_type, frame)

            if isinstance(n1_type, ForwardRef):
                frame = sys._getframe(1)
                n2_type = evaluate_forward_ref(n2_type, frame)

            if cmp_node.children:
                if cmp_node.ch_op == "any":
                    ch_eq = any([ch.eq for ch in cmp_node.children])
                else:
                    ch_eq = all([ch.eq for ch in cmp_node.children] or [True])
            else:
                ch_eq = True
            # check types only of both types are not union
            # otherwise equality was already decided by check above

            if type(n1_type) is TypeVar and type(n2_type) is TypeVar:
                ch_eq &= n1_type == n1_type
            elif type(n1_type) is TypeVar:
                if n2_type == Union:
                    ch_eq = True
                else:
                    ch_eq = False
            elif type(n2_type) is TypeVar:
                if n1_type == Union:
                    ch_eq = True
                else:
                    ch_eq = False
                ch_eq = False

            elif n1_type != Union and n2_type != Union:
                if not inspect.isclass(n1_type):
                    raise TypeError(f"{n1_type} is not class")
                if not inspect.isclass(n2_type):
                    raise TypeError(f"{n2_type} is not class")
                ch_eq &= issubclass(n1_type, n2_type)

            elif n1_type != Union and n2_type == Union:
                if not inspect.isclass(n1_type):
                    raise TypeError(f"{n1_type} is not class")
                ch_eq &= any(
                    [issubclass(n1_type, t) for t in [int, float, str, bool, type(None), ABase]]
                )
            cmp_node.eq = ch_eq

        return root_node.eq

    def to_json(self) -> Dict[str, Any]:
        """Dump TypeNode to json."""
        pre_order: Dict[str, Any] = {"root": {}}
        stack: List[Tuple[ABase, Dict[str, Any], str]] = [(self, pre_order, "root")]
        while stack:
            current, current_parent, parent_key = stack.pop(0)

            type_ = current.type_
            if isinstance(type_, ForwardRef):
                type_ = evaluate_forward_ref(type_, sys._getframe(1))

            if hasattr(type_, "__orig_qualname__"):
                type_name = type_.__orig_qualname__
                module = type_.__module__
            elif hasattr(type_, "__qualname__"):
                type_name = type_.__qualname__
                module = type_.__module__
            elif type_ is None:
                type_name = "NoneType"
                module = "types"
            else:
                type_name = type_.__name__
                module = type_.__module__
            current_parent[parent_key] = {
                "type": type_name,
                "args": [
                    None,
                ]
                * len(current.children),
                "module": module,
            }
            for n, arg in enumerate(current.children):
                stack.append((arg, current_parent[parent_key]["args"], n))

        self._json_cache.append((self.copy(subclass_check=False), pre_order["root"]))
        return pre_order["root"]

    @classmethod
    def from_json(cls, json_data, _locals={}) -> "TypeNode":
        """Load TypeNode from json."""
        root_parent = cls(None)
        root_parent.children = [None]

        stack: List[Tuple[TypeNode, str | int, str, str, Dict[str, Any]]] = [
            (root_parent, 0, json_data["type"], json_data["module"], json_data["args"])
        ]
        pre_order = []

        while stack:
            parent, parent_key, _type, _type_mod, _args = stack.pop(0)
            mod = importlib.import_module(_type_mod)
            _type_path = _type.split(".")
            _type_o = mod
            for path_part in _type_path:
                if path_part == "<locals>":
                    _type_o = _locals
                else:
                    if path_part != "NoneType":
                        if isinstance(_type_o, dict):
                            _type_o = _type_o[path_part]
                        else:
                            _type_o = getattr(_type_o, path_part)
                    else:
                        _type_o = None

            type_node = cls(_type_o)
            type_node.children = [None] * len(_args)
            pre_order.insert(0, (parent, parent_key, type_node))
            for n, arg in enumerate(_args):
                stack.insert(0, (type_node, n, arg["type"], arg["module"], arg["args"]))

        for po_entry in pre_order:
            parent, parent_key, type_node = po_entry
            parent.children[parent_key] = type_node

        return root_parent.children[0]

    def copy(self, subclass_check=True):
        """Return new copy of TypeNode with subclass_check override."""
        root_node = TypeNode(type_=self.type_, subclass_check=subclass_check)
        root_node.children = [None]
        stack = [(self, 0, root_node)]
        while stack:
            current, parent_index, current_parent = stack.pop(0)
            new_current = TypeNode(type_=current.type_, subclass_check=subclass_check)
            new_current.children = [None] * len(current.children)
            for n, ch in enumerate(current.children):
                stack.insert(0, (ch, n, new_current))
            current_parent.children[parent_index] = new_current

        return root_node.children[0]



# flake8: noqa

INPUT_PORTS_GROUP = {
    "position": {"name": "top"},
    "attrs": {"portBody": {"magnet": "passive", "r": 10, "fill": "#023047", "stroke": "#023047"}},
    "label": {
        "position": {"name": "top", "args": {"y": -20, "x": -20, "angle": 45}},
        "markup": [{"tagName": "text", "selector": "label", "className": "label-text"}],
    },
    "markup": [{"tagName": "circle", "selector": "portBody"}],
}
STANDALONE_INPUT_PORTS_GROUP = {
    "position": {"name": "bottom"},
    "attrs": {"portBody": {"magnet": True, "r": 10, "fill": "#023047", "stroke": "#023047"}},
    "label": {
        "position": {
            "name": "right",
            "args": {
                "y": 0,
            },
        },
        "markup": [{"tagName": "text", "selector": "label", "className": "label-text"}],
    },
    "markup": [{"tagName": "circle", "selector": "portBody"}],
}

OUTPUT_PORTS_GROUP = {
    "position": {"name": "bottom"},
    "attrs": {"portBody": {"magnet": True, "r": 10, "fill": "#E6A502", "stroke": "#023047"}},
    "label": {
        "position": {"name": "bottom", "args": {"y": 20, "x": 20, "angle": 45}},
        "markup": [{"tagName": "text", "selector": "label", "className": "label-text"}],
    },
    "markup": [{"tagName": "circle", "selector": "portBody"}],
}

ARGS_PORTS_GROUP = {
    "position": {"name": "left"},
    "attrs": {"portBody": {"magnet": True, "r": 10, "fill": "#11e602", "stroke": "#023047"}},
    "label": {
        "position": {"name": "left", "args": {"y": 6}},
        "markup": [{"tagName": "text", "selector": "label", "className": "label-text"}],
    },
    "markup": [{"tagName": "circle", "selector": "portBody"}],
}

STANDALONE_ARGS_PORTS_GROUP = {
    "position": {"name": "right"},
    "attrs": {"portBody": {"magnet": True, "r": 10, "fill": "#11e602", "stroke": "#023047"}},
    "label": {
        "position": {"name": "right", "args": {"y": 6}},
        "markup": [{"tagName": "text", "selector": "label", "className": "label-text"}],
    },
    "markup": [{"tagName": "circle", "selector": "portBody"}],
}

RESOURCES_PORTS_GROUP = {
    "position": {"name": "left"},
    "attrs": {"portBody": {"magnet": True, "r": 10, "fill": "#bf3b3b", "stroke": "#023047"}},
    "label": {
        "position": {"name": "left", "args": {"y": 0}},
        "markup": [{"tagName": "text", "selector": "label", "className": "label-text"}],
    },
    "markup": [{"tagName": "circle", "selector": "portBody"}],
}

STANDALONE_RESOURCES_PORTS_GROUP = {
    "position": {"name": "right"},
    "attrs": {"portBody": {"magnet": True, "r": 10, "fill": "#bf3b3b", "stroke": "#023047"}},
    "label": {
        "position": {"name": "right", "args": {"y": 0}},
        "markup": [{"tagName": "text", "selector": "label", "className": "label-text"}],
    },
    "markup": [{"tagName": "circle", "selector": "portBody"}],
}


def make_in_ports(traction):
    for f, tf in traction._fields.items():
        if f.startswith("i_"):
            yield {
                "group": "ins",
                "attrs": {
                    "label": {
                        "text": f,
                    }
                },
                "label": {
                    "position": {
                        "name": "top",
                        "args": {"y": -20 - len(f) * 2, "x": -20 - len(f) * 2, "angle": 45},
                    }
                },
                "id": f"{traction.uid}:{f}",
            }


def make_out_ports(traction):
    for f, tf in traction._fields.items():
        if f.startswith("o_"):
            yield {
                "group": "outs",
                "attrs": {"label": {"text": f}},
                "label": {
                    "position": {
                        "name": "top",
                        "args": {"y": 20 + len(f) * 2, "x": 20 + len(f) * 2, "angle": 45},
                    }
                },
                "id": f"{traction.uid}:{f}",
            }


def make_resource_ports(traction):
    for f, tf in traction._fields.items():
        if f.startswith("r_"):
            yield {
                "group": "resources",
                "attrs": {"label": {"text": f}},
                "id": f"{traction.uid}:{f}",
            }


def make_args_ports(traction):
    for f, tf in traction._fields.items():
        if f.startswith("a_"):
            yield {
                "group": "resources",
                "attrs": {
                    "portBody": {"magnet": True, "r": 10, "fill": "#11e602", "stroke": "#023047"},
                    "label": {"text": f},
                },
                "id": f"{traction.uid}:{f}",
            }


def get_traction_args(traction):
    ret = []
    for f, tf in traction._fields.items():
        if f.startswith("a_"):
            ret.append(f)
    return ret


def get_traction_resources(traction):
    ret = []
    for f, tf in traction._fields.items():
        if f.startswith("r_"):
            ret.append(f)
    return ret


def get_traction_inputs(traction):
    ret = []
    for f, tf in traction._fields.items():
        if f.startswith("i_"):
            ret.append(f)
    return ret


def get_traction_outputs(traction):
    ret = []
    for f, tf in traction._fields.items():
        if f.startswith("o_"):
            ret.append(f)
    return ret


def traction_cell(traction, x, y, width, height):
    return {
        "type": "standard.Rectangle",
        "position": {"x": x, "y": y},
        "size": {
            "width": width,
            "height": height,
        },
        "angle": 0,
        "ports": {
            "groups": {
                "ins": INPUT_PORTS_GROUP,
                "outs": OUTPUT_PORTS_GROUP,
                "args": ARGS_PORTS_GROUP,
                "resources": RESOURCES_PORTS_GROUP,
            },
            "items": [
                *make_in_ports(traction),
                *make_out_ports(traction),
                *make_resource_ports(traction),
                *make_args_ports(traction),
            ],
        },
        "id": traction.uid,
        "z": 1,
        "attrs": {
            "body": {"fill": "#8ECAE6"},
            "label": {
                "fontSize": 16,
                "text": traction.__class__.__name__ + "\n" + traction.uid,
                "y": -10,
            },
            "id": traction.__class__.__name__ + ":" + traction.uid,
            "root": {"magnet": False},
        },
    }


def make_standalone_ioar(name, x, y, group):
    return {
        "type": "standard.Rectangle",
        "position": {"x": x, "y": y},
        "size": {"width": 150, "height": 30},
        "angle": 0,
        "ports": {
            "groups": {
                "ins": STANDALONE_INPUT_PORTS_GROUP,
                "outs": OUTPUT_PORTS_GROUP,
                "args": STANDALONE_ARGS_PORTS_GROUP,
                "resources": STANDALONE_RESOURCES_PORTS_GROUP,
            },
            "items": [{"group": group, "attrs": {"label": {}}, "id": name}],
        },
        "id": name,
        "z": 1,
        "attrs": {
            "body": {"fill": "#8ECAE6"},
            "label": {"fontSize": 16, "text": name, "y": 0},
            "id": "#",
            "root": {"magnet": False},
        },
    }


def make_connection(source, source_port, target, target_port):
    return {
        "type": "standard.Link",
        "source": {"id": source, "port": source_port},
        "target": {"id": target, "port": target_port},
        "z": 1,
        "attrs": {"line": {"stroke": "#000000", "strokeWidth": 1}},
    }


def tractor_graph(tractor):
    cells = []
    y = 0
    x = 0
    for f, tf in tractor._fields.items():
        if f.startswith("i_"):
            cells.append(make_standalone_ioar(f, x, y, "ins"))
            x += 200

    y = 200
    x = 0
    for f, tf in tractor._fields.items():
        if f.startswith("r_"):
            cells.append(make_standalone_ioar(f, x, y, "resources"))
            y += 50
        if f.startswith("a_"):
            cells.append(make_standalone_ioar(f, x, y, "args"))
            y += 50

    y = 200
    x = 400
    for f, tf in tractor._fields.items():
        if f.startswith("t_"):
            outputs = get_traction_outputs(getattr(tractor, f))
            inputs = get_traction_inputs(getattr(tractor, f))
            max_out = max([len(x) for x in outputs] or [0]) * 8
            max_in = max([len(x) for x in inputs] or [0]) * 8

            y += max_in
            width = (
                max(len(getattr(tractor, f).uid), len(getattr(tractor, f).__class__.__name__)) + 20
            ) * 10
            height = max(
                (
                    len(get_traction_args(getattr(tractor, f)))
                    + len(get_traction_resources(getattr(tractor, f)))
                )
                * 20,
                70,
            )

            cells.append(traction_cell(getattr(tractor, f), x, y, width, height))
            y += height + max_out
    for (source, source_id), (target, target_id) in tractor._io_map.items():
        print(source, source_id)
        if source != "#":
            source_t = getattr(tractor, source).uid
        else:
            source_t = source_id
        if target != "#":
            target_t = getattr(tractor, target).uid
        else:
            target_t = target_id
        cells.append(
            make_connection(
                source_t, source_t + ":" + source_id, target_t, target_t + ":" + target_id
            )
        )

    return {"cells": cells}



import argparse
from . import container_runner
from . import catalog
from .web import run as web
from . import runner

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Pytraction container ep")
    subparsers = parser.add_subparsers(required=True, dest="command")
    container_runner.make_parsers(subparsers)
    catalog.make_parsers(subparsers)
    web.make_parsers(subparsers)
    runner.make_parsers(subparsers)

    args = parser.parse_args()
    args.command(args)



from pytractions.base import (
    TList,
    Out,
    In,
)

from pytractions.transformations import Flatten, FilterDuplicates, ListMultiplier


def test_flatten():
    t_flatten = Flatten[int](
        uid="test-flatten",
        i_complex=In[TList[TList[int]]](
            data=TList[TList[int]](
                [
                    TList[int]([1, 2]),
                    TList[int]([3, 4]),
                ]
            )
        ),
    )
    t_flatten.run()
    assert t_flatten.o_flat == Out[TList[int]](data=TList[int]([1, 2, 3, 4]))


def test_filter_duplicates():
    t_filter_duplicates = FilterDuplicates[int](
        uid="test-filter-duplicates",
        i_list=In[TList[int]](data=TList[int]([1, 1, 1, 2])),
    )
    t_filter_duplicates.run()
    assert t_filter_duplicates.o_list == Out[TList[int]](data=TList[int]([1, 2]))


def test_list_multiplier():
    t_list_multiplier = ListMultiplier[int, str](
        uid="test-list-multiplier",
        i_list=In[TList[int]](data=TList[int]([1, 1, 1, 1, 1])),
        i_scalar=In[str](data="a"),
    )
    t_list_multiplier.run()
    assert t_list_multiplier.o_list == Out[TList[str]](data=TList[str](["a", "a", "a", "a", "a"]))



from typing import Type, Union

import pytest

from pytractions.base import (
    Traction,
    TList,
    Out,
    In,
    Arg,
    Res,
    Base,
    STMDExecutorType,
    STMD,
    STMDSingleIn,
)


from pytractions.tractor import Tractor


class NOPResource(Base):
    """No operation resource."""

    pass


class NOPResource2(Base):
    """No operation resource2."""

    pass


class EmptyTraction(Traction):
    """Empty traction."""

    i_input: In[int]
    o_output: Out[int]
    a_arg: Arg[int]
    r_res: Res[NOPResource]

    def _run(self, on_update) -> None:  # pragma: no cover
        if not self.i_input.data:
            print(self.uid, "DATA", self.i_input.data)
        if not self.i_coeficient.data:
            print(self.uid, "COEF", self.i_coeficient.data)
        self.o_output.data = self.i_input.data * 2 / self.i_coeficient.data


class Double(Traction):
    """Double the input and divide by coeficient."""

    i_input: In[float]
    i_coeficient: In[float]
    o_output: Out[float]

    def _run(self, on_update) -> None:  # pragma: no cover
        if not self.i_input.data:
            print(self.uid, "DATA", self.i_input.data)
        if not self.i_coeficient.data:
            print(self.uid, "COEF", self.i_coeficient.data)
        self.o_output.data = self.i_input.data * 2 / self.i_coeficient.data


class STMDDouble(STMD):
    """Double the inputs and divide by coeficient."""

    _traction: Type[Traction] = Double

    i_input: In[TList[float]]
    i_coeficient: STMDSingleIn[float]

    o_output: Out[TList[float]]


class Half(Traction):
    """Half the input and multiply by coeficient."""

    i_input: In[float]
    i_coeficient: In[float]
    o_output: Out[float]

    def _run(self, on_update) -> None:  # pragma: no cover
        if not self.i_input.data:
            print(self.uid, "DATA", self.i_input.data)
        if not self.i_coeficient.data:
            print(self.uid, "COEF", self.i_coeficient.data)
        self.o_output.data = self.i_input.data / 2 * self.i_coeficient.data


class STMDHalf(STMD):
    """Half the inputs and multiply by coeficient."""

    _traction: Type[Traction] = Half

    i_input: In[TList[float]]
    i_coeficient: STMDSingleIn[float]
    o_output: Out[TList[float]]


class Calculator(Tractor):
    """Calculator."""

    i_inputs: In[TList[float]] = In[TList[float]]()
    i_coeficient: STMDSingleIn[float] = STMDSingleIn[float]()
    a_pool_size: Arg[int] = Arg[int](a=30)

    t_double: STMDDouble = STMDDouble(
        uid="double",
        i_input=i_inputs,
        i_coeficient=i_coeficient,
        a_executor_type=Arg[STMDExecutorType](a=STMDExecutorType.THREAD),
        a_pool_size=a_pool_size,
    )

    t_half: STMDHalf = STMDHalf(
        uid="half",
        i_input=t_double.o_output,
        i_coeficient=i_coeficient,
        a_executor_type=Arg[STMDExecutorType](a=STMDExecutorType.THREAD),
        a_pool_size=a_pool_size,
    )

    o_output: Out[TList[float]] = t_half.o_output


def test_stmd_attr_validation():
    # wrong input
    with pytest.raises(TypeError):

        class TestSTMD1(STMD):
            _traction: Type[Traction] = EmptyTraction
            i_input: In[int]
            o_output: Out[TList[Out[int]]]
            r_res: Res[NOPResource]
            a_arg: Arg[int]

    # wrong output
    with pytest.raises(TypeError):

        class TestSTMD2(STMD):
            _traction: Type[Traction] = EmptyTraction
            i_input: In[TList[In[int]]]
            o_output: TList[Out[int]]
            r_res: Res[NOPResource]
            a_arg: Arg[int]

    # wrong resource
    with pytest.raises(TypeError):

        class TestSTMD3(STMD):
            _traction: Type[Traction] = EmptyTraction
            i_input: In[TList[In[int]]]
            o_output: Out[TList[Out[int]]]
            r_res: Arg[NOPResource]
            a_arg: Arg[int]

    # wrong resource
    with pytest.raises(TypeError):

        class TestSTMD4(STMD):
            _traction: Type[Traction] = EmptyTraction
            i_input: In[TList[In[int]]]
            o_output: Out[TList[Out[int]]]
            r_res: Res[NOPResource]
            a_arg: In[str]

    # wrong doc arg
    with pytest.raises(TypeError):

        class TestSTMD5(STMD):
            _traction: Type[Traction] = EmptyTraction
            i_input: In[TList[In[int]]]
            o_output: Out[TList[Out[int]]]
            r_res: Res[NOPResource]
            a_arg: Arg[str]
            d_: int

    # wrong attr doc arg
    with pytest.raises(TypeError):

        class TestSTMD6(STMD):
            _traction: Type[Traction] = EmptyTraction
            i_input: In[TList[In[int]]]
            o_output: Out[TList[Out[int]]]
            r_res: Res[NOPResource]
            a_arg: Arg[str]
            d_a_arg: int = 0

    # uknown doc attr
    with pytest.raises(TypeError):

        class TestSTMD7(STMD):
            _traction: Type[Traction] = EmptyTraction
            i_input: In[TList[In[int]]]
            o_output: Out[TList[Out[int]]]
            r_res: Res[NOPResource]
            a_arg: Arg[str]
            d_a_uknown: str = ""

    # custom attr
    with pytest.raises(TypeError):

        class TestSTMD(STMD):
            _traction: Type[Traction] = EmptyTraction
            i_input: In[TList[In[int]]]
            o_output: Out[TList[Out[int]]]
            r_res: Res[NOPResource]
            a_arg: Arg[str]
            custom: str

    # wrong inner type
    with pytest.raises(TypeError):

        class TestSTMD8(STMD):
            _traction: Type[Traction] = EmptyTraction
            i_input: In[TList[str]]
            o_output: Out[TList[Out[int]]]
            r_res: Res[NOPResource]
            a_arg: Arg[str]
            custom: str


def test_stmd_calculator():
    c = Calculator(
        uid="calculator",
        i_inputs=In[TList[float]](data=TList[float]([float(x) for x in range(1, 10)])),
        i_coeficient=STMDSingleIn[float](data=0.5),
    )
    c.run()


class G_TTest1(Traction):
    """Test traction."""

    o_out1: Out[float]
    i_in1: In[Union[float, int]]
    a_multiplier: Arg[float]

    def _run(self, on_update) -> None:  # pragma: no cover
        self.o_out1.data = self.i_in1.data * self.a_multiplier.a


class G_TestTractor(Tractor):
    """Test tractor."""

    i_in1: In[Union[float, int]] = In[Union[float, int]](data=20.0)
    a_multiplier: Arg[float] = Arg[float](a=1.0)

    t_traction_1: G_TTest1 = G_TTest1(uid="1", i_in1=i_in1, a_multiplier=a_multiplier)
    t_traction_2: G_TTest1 = G_TTest1(uid="1", i_in1=t_traction_1.o_out1, a_multiplier=a_multiplier)

    o_out1: Out[float] = t_traction_2.o_out1


def test_stmd_tractor(fixture_isodate_now) -> None:
    class TestSTMD(STMD):
        a_multiplier: Arg[float] = Arg[float](a=0.0)
        i_in1: In[TList[Union[float, int]]] = In[TList[Union[float, int]]](
            data=TList[Union[float, int]]([])
        )

        _traction: Type[Traction] = G_TestTractor

        o_out1: Out[TList[float]] = Out[TList[float]](data=TList[float]([]))

    stmd_in1 = In[TList[float]](data=TList[float]([1.0, 2.0, 3.0, 4.0, 5.0]))
    stmd1 = TestSTMD(
        uid="tt1", a_pool_size=Arg[int](a=1), a_multiplier=Arg[float](a=10.0), i_in1=stmd_in1
    )
    stmd1.run()
    assert stmd1.o_out1.data[0] == 100.0
    assert stmd1.o_out1.data[1] == 200.0
    assert stmd1.o_out1.data[2] == 300.0
    assert stmd1.o_out1.data[3] == 400.0
    assert stmd1.o_out1.data[4] == 500.0


def test_stmd_local(fixture_isodate_now) -> None:

    class TestSTMD(STMD):
        a_multiplier: Arg[float] = Arg[float](a=0.0)

        _traction: Type[Traction] = G_TTest1

        o_out1: Out[TList[float]] = Out[TList[float]](data=TList[float]([]))
        i_in1: In[TList[float]]

    stmd_in1 = In[TList[float]](data=TList[float]([1.0, 2.0, 3.0, 4.0, 5.0]))
    stmd1 = TestSTMD(
        uid="tt1", a_pool_size=Arg[int](a=1), a_multiplier=Arg[float](a=10.0), i_in1=stmd_in1
    )
    stmd1.run()
    assert stmd1.o_out1.data[0] == 10.0
    assert stmd1.o_out1.data[1] == 20.0
    assert stmd1.o_out1.data[2] == 30.0
    assert stmd1.o_out1.data[3] == 40.0
    assert stmd1.o_out1.data[4] == 50.0


def test_stmd_threads(fixture_isodate_now) -> None:

    class TestSTMD(STMD):
        a_multiplier: Arg[float] = Arg[float](a=0.0)

        _traction: Type[Traction] = G_TTest1

        o_out1: Out[TList[float]] = Out[TList[float]](data=TList[float]([]))
        i_in1: In[TList[float]]

    stmd_in1 = In[TList[float]](
        data=TList[float](
            [
                1.0,
                2.0,
                3.0,
                4.0,
                5.0,
            ]
        )
    )
    stmd1 = TestSTMD(
        uid="tt1",
        a_executor_type=Arg[STMDExecutorType](a=STMDExecutorType.THREAD),
        a_pool_size=Arg[int](a=1),
        a_multiplier=Arg[float](a=10.0),
        i_in1=stmd_in1,
    )
    stmd1.run()
    assert stmd1.o_out1.data[0] == 10.0
    assert stmd1.o_out1.data[1] == 20.0
    assert stmd1.o_out1.data[2] == 30.0
    assert stmd1.o_out1.data[3] == 40.0
    assert stmd1.o_out1.data[4] == 50.0


class GTestSTMD(STMD):
    """Test STMD."""

    a_multiplier: Arg[float] = Arg[float](a=0.0)

    _traction: Type[Traction] = G_TTest1

    o_out1: Out[TList[float]] = Out[TList[float]](data=TList[float]([]))
    i_in1: In[TList[float]]


def test_stmd_processes(fixture_isodate_now) -> None:

    stmd_in1 = In[TList[float]](
        data=TList[float](
            [
                1.0,
                2.0,
                3.0,
                4.0,
                5.0,
            ]
        )
    )
    stmd1 = GTestSTMD(
        uid="tt1",
        a_executor_type=Arg[STMDExecutorType](a=STMDExecutorType.PROCESS),
        a_pool_size=Arg[int](a=2),
        a_multiplier=Arg[float](a=10.0),
        i_in1=stmd_in1,
    )
    stmd1.run()
    assert stmd1.o_out1.data[0] == 10.0
    assert stmd1.o_out1.data[1] == 20.0
    assert stmd1.o_out1.data[2] == 30.0
    assert stmd1.o_out1.data[3] == 40.0
    assert stmd1.o_out1.data[4] == 50.0



from .base import Traction, STMD
from .tractor import Tractor, MultiTractor

import inspect


def traction_entry_points(module):
    """Discover traction entry points in given module."""
    for k in dir(module):
        v = getattr(module, k)
        if not inspect.isclass(v):
            continue
        if v in (Traction, Tractor, STMD, MultiTractor):
            continue

        if issubclass(v, Traction):
            yield f"{k} = {v.__module__}:{k}"



import enum
import dataclasses
import importlib
from typing import _UnionGenericAlias, Any, Optional, Dict, Type, Tuple, List

from lark import Lark, Transformer
import yaml

from .base import TypeNode, ANY, TList, TDict, JSON_COMPATIBLE, Base


class StrUnion(yaml.YAMLObject):
    """Class for different yaml string formatting."""

    yaml_tag = "OneOf"

    def __init__(self, _list):
        """Initialize StrUnion."""
        self.choices = _list

    def __getitem__(self, n):
        """Get item from choices."""
        return self.choices[n]

    def __setitem__(self, n, x):
        """Set item in choices."""
        self.choices[n] = x


class StrParam:
    """Class for different yaml string formatting."""

    def __init__(self, strparam):
        """Initialize StrParam."""
        self.strparam = strparam


def str_presenter(dumper, data):
    """Yaml string representer with indent."""
    if len(data.splitlines()) > 1:  # check for multiline string
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


def str_param(dumper, data):
    """Yaml str param representer with indent."""
    return dumper.represent_scalar("tag:yaml.org,2002:str", data.strparam, style="|")


def str_multichoice(dumper, data):
    """Yaml str param representer with indent."""
    return dumper.represent_scalar("tag:yaml.org,2002:str", data.strparam, style="|")


yaml.add_representer(str, str_presenter)
yaml.add_representer(StrParam, str_param)


traction_str_grammar = r"""
start: type
type: fullname | fullname "[" typelist "]"
fullname: name ":" name
typelist: type | type "," typelist
name: NAME
NAME: /([a-zA-Z_])+([a-zA-Z_\.])*/
"""


class TTransformer(Transformer):
    """Lark grammar transformer.

    This class is used to by lark parser to import and return class from provided string input
    """

    def __init___(self, *args, **kwargs):
        """Tranformer initializer."""
        super().__init__(*args, **kwargs)
        self.stack = []
        self.current_traction = {"mod": None, "name": None, "args": []}

    def type(self, token):
        """Return generic subclass of class in token."""
        if len(token) > 1:
            return token[0].__class_getitem__(tuple(token[1]))
        else:
            return token[0]

    def typelist(self, token):
        """Handle typelist token."""
        if len(token) > 1:
            return tuple(token)
        else:
            return token

    def name(self, token):
        """Extract name from the parsed token."""
        return str(token[0])

    def fullname(self, token):
        """Import class from module fullname and return class."""
        mod, clsname = token
        mod_ = importlib.import_module(mod)
        return getattr(mod_, clsname)

    def start(self, token):
        """Handle grammer start token."""
        return token[0]


def parse_traction_str(traction_str):
    """Parse traction string and return traction class."""
    parser = Lark(traction_str_grammar, parser="lalr", transformer=TTransformer())
    parsed = parser.parse(traction_str)
    return parsed


def generate_type_description(type_, indent=0):
    """Generate description for provided type."""
    if type_ in (str, int, float, bool, type(None)):
        return f"{type_.__name__}"
    elif type_.__class__ == _UnionGenericAlias:
        return f"Optional[{generate_type_description(type_.__args__[0], indent=indent)}]"
    elif TypeNode.from_type(type_) == TypeNode.from_type(TList[ANY]):
        return f"List[{generate_type_description(type_._params[0], indent=indent)}]"
    elif TypeNode.from_type(type_) == TypeNode.from_type(TDict[ANY, ANY]):
        return (
            f"Dict[{generate_type_description(type_._params[0], indent=indent)},"
            f"{generate_type_description(type_._params[1], indent=indent)}]"
        )
    elif issubclass(type_, enum.Enum):
        return f"{type_.__name__}"
    else:
        fields = ",\n".join(
            [
                " " * (indent + 4) + f"{f}: {generate_type_description(t, indent=indent+4)}"
                for f, t in type_._fields.items()
                if not f.startswith("_")
            ]
        )
        ret = f"{type_.__name__}(\n{fields}\n)"
    return ret


def generate_ari_description(traction, field):
    """Generate description for arg, resource or input field of traction."""
    description = f"""DESCRIPTION: '{getattr(traction, "d_" + field, None) or ""}'

TYPE: {generate_type_description(traction._fields[field])}"""
    return description


def generate_traction_ari(traction):
    """Generate traction args, resources and inputs."""
    args = []
    inputs = []
    resources = []
    for f, fv in traction._fields.items():
        if f.startswith("a_"):
            arg = {"name": f, "description": generate_ari_description(traction, f)}
            args.append(arg)
        elif f.startswith("i_"):
            input_ = {"name": f, "description": generate_ari_description(traction, f)}
            inputs.append(input_)
        elif f.startswith("r_"):
            resource = {"name": f, "description": generate_ari_description(traction, f)}
            resources.append(resource)
    ari_str = yaml.dump_all(
        [
            {"name": p["name"], "data": StrParam(f'{p["description"]}')}
            for p in args + resources + inputs
        ]
    )
    return ari_str


def get_traction_defaults(traction):
    """Get defaults for all traction attributes."""
    defaults = {}
    for f in traction._fields:
        if hasattr(traction, f):
            tf = getattr(traction, f, dataclasses._MISSING_TYPE)
        else:
            tf = traction.__dataclass_fields__[f].default
        if f.startswith("a_"):
            if not isinstance(tf, dataclasses._MISSING_TYPE):
                defaults[f] = StrParam(yaml.dump(tf.content_to_json()))
    if f.startswith("i_"):
        if not isinstance(tf, dataclasses._MISSING_TYPE):
            defaults[f] = StrParam(yaml.dump(tf.content_to_json()))
        if f.startswith("r_"):
            if not isinstance(tf, dataclasses._MISSING_TYPE):
                defaults[f] = StrParam(yaml.dump(tf.content_to_json()))
    return defaults


def gen_default_inputs(traction_cls):
    """Get defaults for all traction attributes."""
    defaults = {str: "string", int: 12345, float: 1.2345, None: None, bool: False, type(None): None}
    pre_order: Dict[str, Any] = {}
    # stack is list of (current_cls_to_process, current_parent, current_key, current_default)
    stack: List[Tuple[Type[Base], Dict[str, Any], str, Optional[JSON_COMPATIBLE]]] = []
    current = traction_cls
    current_parent = pre_order
    parent_key = "root"
    current_default = None

    current_parent[parent_key] = {}
    for f, ftype in current._fields.items():
        if not (f.startswith("a_") or f.startswith("i_") or f.startswith("r_")):
            continue
        if type(current.__dataclass_fields__[f].default) in (str, int, float, None, bool):
            stack.append(
                (
                    ftype,
                    current_parent[parent_key],
                    f,
                    current.__dataclass_fields__[f].default,
                )
            )
        else:
            stack.append((ftype, current_parent[parent_key], f, None))

    while stack:
        current, current_parent, parent_key, current_default = stack.pop(0)
        if current.__class__ == _UnionGenericAlias:  # and current._name != "Optional":
            current_parent[parent_key] = StrUnion([None] * len(current.__args__))
            for n, uarg in enumerate(current.__args__):
                if type(uarg) in (str, int, float, None, bool):
                    stack.append(
                        (
                            uarg,
                            current_parent[parent_key],
                            n,
                            uarg,
                        )
                    )
                else:
                    stack.append((uarg, current_parent[parent_key], n, None))
        elif issubclass(current, enum.Enum):
            current_parent[parent_key] = list(current.__members__.keys())
        elif TypeNode.from_type(current) == TypeNode.from_type(TList[ANY]):
            current_parent[parent_key] = [None]
            stack.append((current._params[0], current_parent[parent_key], 0, None))

        elif hasattr(current, "_fields"):
            current_parent[parent_key] = {}
            for f, ftype in current._fields.items():
                if type(current.__dataclass_fields__[f].default) in (str, int, float, None, bool):
                    stack.append(
                        (
                            ftype,
                            current_parent[parent_key],
                            f,
                            current.__dataclass_fields__[f].default,
                        )
                    )
                else:
                    stack.append((ftype, current_parent[parent_key], f, None))
        else:
            current_parent[parent_key] = (
                defaults[current] if current_default is None else current_default
            )
    return pre_order["root"]



from typing import Union, TypeVar
from dataclasses import field

import pytest

from pytractions.base import (
    Traction,
    TList,
    Out,
    In,
    Arg,
    Res,
    Base,
    NoData,
    TypeNode,
)


from pytractions.tractor import Tractor

T = TypeVar("T")


def test_traction_ok_args_1():
    class TTest(Traction):
        i_in1: In[int]
        o_out1: Out[int]
        r_res1: Res[int]
        a_arg1: Arg[int]


def test_traction_ok_args_description_1():
    class TTest(Traction):
        i_in1: In[int]
        o_out1: Out[int]
        r_res1: Res[int]
        a_arg1: Arg[int]
        d_i_in1: str = "Description of i_in1"


def test_traction_wrong_args_1():
    with pytest.raises(TypeError):

        class TTest(Traction):
            i_in1: Out[int]


def test_traction_wrong_args_2():
    with pytest.raises(TypeError):

        class TTest(Traction):
            o_out1: In[int]


def test_traction_wrong_args_3():
    with pytest.raises(TypeError):

        class TTest(Traction):
            a_arg1: In[int]


def test_traction_wrong_args_4():
    with pytest.raises(TypeError):

        class TTest(Traction):
            r_res1: In[int]


def test_traction_inputs_1():
    class TTest(Traction):
        i_in1: In[int] = In[int]()

    o: Out[int] = Out[int](data=10)
    TTest(uid="1", i_in1=o)


def test_traction_inputs_read():
    class TTest(Traction):
        i_in1: In[int] = field(default=In[int]())

    o: Out[int] = Out[int](data=10)
    t = TTest(uid="1", i_in1=o)
    assert id(t.i_in1.data) == id(o.data)


def test_traction_inputs_read_unset():
    class TTest(Traction):
        i_in1: In[int]

    t = TTest(uid="1")
    assert TypeNode.from_type(type(t.i_in1)) == TypeNode.from_type(NoData[int])


def test_traction_inputs_read_set():
    class TTest(Traction):
        i_in1: In[int]

    o: Out[int] = Out[int](data=10)
    t = TTest(uid="1", i_in1=o)
    assert TypeNode.from_type(type(t.i_in1)) == TypeNode.from_type(Out[int])
    assert t.i_in1.data == 10


def test_traction_to_json():
    class TTest(Traction):
        i_in1: In[int] = In[int]()
        d_i_in1: str = "description of i_in1"

    o: Out[int] = Out[int](data=10)
    t = TTest(uid="1", i_in1=o)
    assert t.to_json() == {
        "$data": {
            "d_i_in1": "description of i_in1",
            "details": {
                "$data": {},
                "$type": {
                    "args": [
                        {"args": [], "type": "str", "module": "builtins"},
                        {"args": [], "type": "str", "module": "builtins"},
                    ],
                    "type": "TDict",
                    "module": "pytractions.base",
                },
            },
            "errors": {
                "$data": [],
                "$type": {
                    "args": [{"args": [], "type": "str", "module": "builtins"}],
                    "type": "TList",
                    "module": "pytractions.base",
                },
            },
            "i_in1": {
                "$data": {"data": 10},
                "$type": {
                    "args": [{"args": [], "type": "int", "module": "builtins"}],
                    "type": "Out",
                    "module": "pytractions.base",
                },
            },
            "skip": False,
            "skip_reason": "",
            "state": "ready",
            "stats": {
                "$data": {"finished": "", "skipped": False, "started": ""},
                "$type": {"args": [], "type": "TractionStats", "module": "pytractions.base"},
            },
            "uid": "1",
        },
        "$type": {
            "args": [],
            "module": "tests.test_base_traction",
            "type": "test_traction_to_json.<locals>.TTest",
        },
    }


def test_traction_inlist_to_json():
    class TTest(Traction):
        i_in1: In[TList[int]]

    print(TTest._fields)

    o: Out[TList[int]] = Out[TList[int]](data=TList[int]([10]))
    t = TTest(uid="1", i_in1=o)
    assert t.to_json() == {
        "$data": {
            "details": {
                "$data": {},
                "$type": {
                    "args": [
                        {"args": [], "type": "str", "module": "builtins"},
                        {"args": [], "type": "str", "module": "builtins"},
                    ],
                    "type": "TDict",
                    "module": "pytractions.base",
                },
            },
            "errors": {
                "$data": [],
                "$type": {
                    "args": [{"args": [], "type": "str", "module": "builtins"}],
                    "type": "TList",
                    "module": "pytractions.base",
                },
            },
            "i_in1": {
                "$data": {
                    "data": {
                        "$data": [10],
                        "$type": {
                            "args": [{"args": [], "module": "builtins", "type": "int"}],
                            "module": "pytractions.base",
                            "type": "TList",
                        },
                    }
                },
                "$type": {
                    "args": [
                        {
                            "args": [{"args": [], "module": "builtins", "type": "int"}],
                            "module": "pytractions.base",
                            "type": "TList",
                        }
                    ],
                    "module": "pytractions.base",
                    "type": "Out",
                },
            },
            "skip": False,
            "skip_reason": "",
            "state": "ready",
            "stats": {
                "$data": {"finished": "", "skipped": False, "started": ""},
                "$type": {"args": [], "type": "TractionStats", "module": "pytractions.base"},
            },
            "uid": "1",
        },
        "$type": {
            "args": [],
            "module": "tests.test_base_traction",
            "type": "test_traction_inlist_to_json.<locals>.TTest",
        },
    }


def test_traction_to_from_json():
    class TTest(Traction):
        i_in1: In[int]  # = In[int]()

    o: Out[int] = Out[int](data=10)
    t = TTest(uid="1", i_in1=o)
    t2 = TTest.from_json(t.to_json(), _locals=locals())
    assert t == t2


def test_traction_outputs_no_init():
    class TTest(Traction):
        o_out1: Out[int]

    t = TTest(uid="1")
    assert t.o_out1 == Out[int](data=int())


def test_traction_outputs_no_init_custom_default():
    class TTest(Traction):
        o_out1: Out[int] = Out[int](data=10)

    t = TTest(uid="1")
    assert t.o_out1 == Out[int](data=10)


def test_traction_outputs_uid():
    class TTest(Traction):
        o_out1: Out[int] = Out[int](data=10)

    t = TTest(uid="1")
    assert t.o_out1.uid == "TTest[1]::o_out1"


def test_traction_chain():
    class TTest1(Traction):
        o_out1: Out[int]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1.data = 20

    class TTest2(Traction):
        i_in1: In[int]
        o_out1: Out[int]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1.data = self.i_in1.data + 10

    t1 = TTest1(uid="1")
    t2 = TTest2(uid="1", i_in1=t1.o_out1)

    t1.run()
    t2.run()
    assert t2.o_out1.data == 30


def test_traction_chain_in_to_out():
    class TTest1(Traction):
        o_out1: Out[int]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1.data = 20

    class TTest2(Traction):
        i_in1: In[int]
        o_out1: Out[int]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1 = self.i_in1

    t1 = TTest1(uid="1")
    t2 = TTest2(uid="1", i_in1=t1.o_out1)

    t1.run()
    t2.run()
    assert t2.o_out1.data == 20
    t1.o_out1.data = 30

    assert t2.i_in1.data == 30


def test_traction_json(fixture_isodate_now):
    class TTest1(Traction):
        o_out1: Out[int]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1.data = 20

    class TTest2(Traction):
        i_in1: In[Union[int, float]]
        o_out1: Out[int]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1.data = self.i_in1.data + 10

    t1 = TTest1(uid="1")
    t2 = TTest2(uid="2", i_in1=t1.o_out1)

    t1.run()
    t2.run()
    assert t1.to_json() == {
        "$data": {
            "details": {
                "$data": {},
                "$type": {
                    "args": [
                        {"args": [], "type": "str", "module": "builtins"},
                        {"args": [], "type": "str", "module": "builtins"},
                    ],
                    "type": "TDict",
                    "module": "pytractions.base",
                },
            },
            "errors": {
                "$data": [],
                "$type": {
                    "args": [{"args": [], "type": "str", "module": "builtins"}],
                    "type": "TList",
                    "module": "pytractions.base",
                },
            },
            "o_out1": {
                "$data": {"data": 20},
                "$type": {
                    "args": [{"args": [], "type": "int", "module": "builtins"}],
                    "type": "Out",
                    "module": "pytractions.base",
                },
            },
            "skip": False,
            "skip_reason": "",
            "state": "finished",
            "stats": {
                "$data": {
                    "finished": "1990-01-01T00:00:01.00000Z",
                    "skipped": False,
                    "started": "1990-01-01T00:00:00.00000Z",
                },
                "$type": {"args": [], "module": "pytractions.base", "type": "TractionStats"},
            },
            "uid": "1",
        },
        "$type": {
            "args": [],
            "module": "tests.test_base_traction",
            "type": "test_traction_json.<locals>.TTest1",
        },
    }

    assert t2.to_json() == {
        "$data": {
            "details": {
                "$data": {},
                "$type": {
                    "args": [
                        {"args": [], "type": "str", "module": "builtins"},
                        {"args": [], "type": "str", "module": "builtins"},
                    ],
                    "type": "TDict",
                    "module": "pytractions.base",
                },
            },
            "errors": {
                "$data": [],
                "$type": {
                    "args": [{"args": [], "type": "str", "module": "builtins"}],
                    "type": "TList",
                    "module": "pytractions.base",
                },
            },
            "i_in1": "TTest1[1]#o_out1",
            "o_out1": {
                "$data": {"data": 30},
                "$type": {
                    "args": [{"args": [], "type": "int", "module": "builtins"}],
                    "type": "Out",
                    "module": "pytractions.base",
                },
            },
            "skip": False,
            "skip_reason": "",
            "state": "finished",
            "stats": {
                "$data": {
                    "finished": "1990-01-01T00:00:03.00000Z",
                    "skipped": False,
                    "started": "1990-01-01T00:00:02.00000Z",
                },
                "$type": {"args": [], "module": "pytractions.base", "type": "TractionStats"},
            },
            "uid": "2",
        },
        "$type": {
            "args": [],
            "module": "tests.test_base_traction",
            "type": "test_traction_json.<locals>.TTest2",
        },
    }


def test_tractor_members_order() -> None:
    class TTest1(Traction):
        o_out1: Out[float]
        a_multiplier: Arg[float]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1.data = 20 * self.a_multiplier.a

    class TTest2(Traction):
        i_in1: In[float]
        o_out1: Out[float]
        a_reducer: Arg[float]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1.data = (self.i_in1.data + 10) / float(self.a_reducer.a)

    class TestTractor(Tractor):
        a_multiplier: Arg[float] = Arg[float](a=0.0)
        a_reducer: Arg[float] = Arg[float](a=0.0)

        t_ttest1: TTest1 = TTest1(uid="1", a_multiplier=a_multiplier)
        t_ttest4: TTest2 = TTest2(uid="4", a_reducer=a_reducer)
        t_ttest3: TTest2 = TTest2(uid="3", a_reducer=a_reducer)
        t_ttest2: TTest2 = TTest2(uid="2", a_reducer=a_reducer)

        t_ttest2.i_in1 = t_ttest1.o_out1
        t_ttest3.i_in1 = t_ttest4.o_out1
        t_ttest4.i_in1 = t_ttest1.o_out1

        o_out1: Out[float] = t_ttest4.o_out1

    ttrac = TestTractor(uid="t1")

    tractions = []
    for k, v in ttrac.__dict__.items():
        if k.startswith("t_"):
            tractions.append(k)

    assert tractions == ["t_ttest1", "t_ttest4", "t_ttest3", "t_ttest2"]


def test_tractor_members_invalid_order() -> None:
    class TTest1(Traction):
        o_out1: Out[float]
        a_multiplier: Arg[float]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1.data = 20 * self.a_multiplier.a

    class TTest2(Traction):
        i_in1: In[float]
        o_out1: Out[float]
        a_reducer: Arg[float]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1.data = (self.i_in1.data + 10) / float(self.a_reducer.a)

    with pytest.raises(ValueError):

        class TestTractor(Tractor):
            a_multiplier: Arg[float] = Arg[float](a=0.0)
            a_reducer: Arg[float] = Arg[float](a=0.0)

            t_ttest1: TTest1 = TTest1(uid="1", a_multiplier=a_multiplier)
            t_ttest2: TTest2 = TTest2(uid="4", a_reducer=a_reducer)
            t_ttest4: TTest2 = TTest2(uid="3", a_reducer=a_reducer)
            t_ttest3: TTest2 = TTest2(uid="2", a_reducer=a_reducer)

            t_ttest2.i_in1 = t_ttest1.o_out1
            t_ttest4.i_in1 = t_ttest3.o_out1

            o_out1: Out[float] = t_ttest4.o_out1


def test_tractor_run() -> None:
    class TTest2(Traction):
        i_in1: In[float]
        o_out1: Out[float]
        a_reducer: Arg[float]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1.data = (self.i_in1.data + 1) / float(self.a_reducer.a)

    class TestTractor(Tractor):
        a_multiplier: Arg[float] = Arg[float](a=0.0)
        a_reducer: Arg[float] = Arg[float](a=0.0)

        i_in1: In[float] = In[float](data=1.0)

        t_ttest1: TTest2 = TTest2(uid="1", a_reducer=a_reducer)
        t_ttest2: TTest2 = TTest2(uid="2", a_reducer=a_reducer)
        t_ttest3: TTest2 = TTest2(uid="3", a_reducer=a_reducer)
        t_ttest4: TTest2 = TTest2(uid="4", a_reducer=a_reducer)

        t_ttest1.i_in1 = i_in1
        t_ttest2.i_in1 = t_ttest1.o_out1
        t_ttest3.i_in1 = t_ttest2.o_out1
        t_ttest4.i_in1 = t_ttest3.o_out1

        o_out1: Out[float] = t_ttest4.o_out1

    tt = TestTractor(
        uid="tt1",
        a_multiplier=Arg[float](a=10.0),
        a_reducer=Arg[float](a=2.0),
        i_in1=In[float](data=10.0),
    )
    tt.run()
    assert tt.o_out1.data == 1.5625


class UselessResource(Base):
    """Testing resource."""

    values_stack: TList[int]

    def get_some_value(self) -> int:
        """Get first value from the init list."""
        return self.values_stack.pop(0)


def test_tractor_run_resources() -> None:

    class TTest2(Traction):
        i_in1: In[float]
        o_out1: Out[float]
        a_reducer: Arg[float]
        r_useless_res: Res[UselessResource]

        def _run(self, on_update) -> None:  # pragma: no cover
            useless_value = self.r_useless_res.r.get_some_value()
            self.o_out1.data = (self.i_in1.data + useless_value) / float(self.a_reducer.a)

    class TestTractor(Tractor):
        a_multiplier: Arg[float] = Arg[float](a=0.0)
        a_reducer: Arg[float] = Arg[float](a=0.0)

        r_useless: Res[UselessResource] = Res[UselessResource](
            r=UselessResource(values_stack=TList[int](TList[int]([])))
        )

        i_in1: In[float] = In[float](data=1.0)

        t_ttest1: TTest2 = TTest2(uid="1", a_reducer=a_reducer, r_useless_res=r_useless)
        t_ttest2: TTest2 = TTest2(uid="2", a_reducer=a_reducer, r_useless_res=r_useless)
        t_ttest3: TTest2 = TTest2(uid="3", a_reducer=a_reducer, r_useless_res=r_useless)
        t_ttest4: TTest2 = TTest2(uid="4", a_reducer=a_reducer, r_useless_res=r_useless)

        t_ttest1.i_in1 = i_in1
        t_ttest2.i_in1 = t_ttest1.o_out1
        t_ttest3.i_in1 = t_ttest2.o_out1
        t_ttest4.i_in1 = t_ttest3.o_out1

        o_out1: Out[float] = t_ttest4.o_out1

    tt = TestTractor(
        uid="tt1",
        a_multiplier=Arg[float](a=10.0),
        a_reducer=Arg[float](a=2.0),
        i_in1=In[float](data=10.0),
        r_useless=Res[UselessResource](r=UselessResource(values_stack=TList[int]([2, 5, 1, 7, 2]))),
    )

    tt.run()
    assert tt.o_out1.data == 5.125


def test_tractor_to_json(fixture_isodate_now) -> None:
    class TTest1(Traction):
        o_out1: Out[float]
        a_multiplier: Arg[float]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1.data = 1 * self.a_multiplier.a

    class TTest2(Traction):
        i_in1: In[float]
        o_out1: Out[float]
        a_reducer: Arg[float]

        def _run(self, on_update) -> None:  # pragma: no cover
            self.o_out1.data = (self.i_in1.data + 1) / float(self.a_reducer.a)

    class TestTractor(Tractor):
        a_multiplier: Arg[float] = Arg[float](a=0.0)
        a_reducer: Arg[float] = Arg[float](a=0.0)

        t_ttest1: TTest1 = TTest1(uid="1", a_multiplier=a_multiplier)
        t_ttest2: TTest2 = TTest2(uid="2", a_reducer=a_reducer)
        t_ttest3: TTest2 = TTest2(uid="3", a_reducer=a_reducer)
        t_ttest4: TTest2 = TTest2(uid="4", a_reducer=a_reducer)

        t_ttest2.i_in1 = t_ttest1.o_out1
        t_ttest3.i_in1 = t_ttest2.o_out1
        t_ttest4.i_in1 = t_ttest3.o_out1

        o_out1: Out[float] = t_ttest4.o_out1

    tt = TestTractor(uid="tt1", a_multiplier=Arg[float](a=10.0), a_reducer=Arg[float](a=2.0))

    tt.run()
    assert tt.to_json() == {
        "$data": {
            "a_multiplier": {
                "$data": {"a": 10.0},
                "$type": {
                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                    "type": "Arg",
                    "module": "pytractions.base",
                },
            },
            "a_reducer": {
                "$data": {"a": 2.0},
                "$type": {
                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                    "type": "Arg",
                    "module": "pytractions.base",
                },
            },
            "details": {
                "$data": {},
                "$type": {
                    "args": [
                        {"args": [], "type": "str", "module": "builtins"},
                        {"args": [], "type": "str", "module": "builtins"},
                    ],
                    "type": "TDict",
                    "module": "pytractions.base",
                },
            },
            "errors": {
                "$data": [],
                "$type": {
                    "args": [{"args": [], "type": "str", "module": "builtins"}],
                    "type": "TList",
                    "module": "pytractions.base",
                },
            },
            "o_out1": {
                "$data": {"data": 2.125},
                "$type": {
                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                    "type": "Out",
                    "module": "pytractions.base",
                },
            },
            "skip": False,
            "skip_reason": "",
            "state": "finished",
            "stats": {
                "$data": {
                    "finished": "1990-01-01T00:00:08.00000Z",
                    "skipped": False,
                    "started": "1990-01-01T00:00:00.00000Z",
                },
                "$type": {"args": [], "module": "pytractions.base", "type": "TractionStats"},
            },
            "t_ttest1": {
                "$data": {
                    "a_multiplier": {
                        "$data": {"a": 0.0},
                        "$type": {
                            "args": [{"args": [], "type": "float", "module": "builtins"}],
                            "type": "Arg",
                            "module": "pytractions.base",
                        },
                    },
                    "details": {
                        "$data": {},
                        "$type": {
                            "args": [
                                {"args": [], "type": "str", "module": "builtins"},
                                {"args": [], "type": "str", "module": "builtins"},
                            ],
                            "type": "TDict",
                            "module": "pytractions.base",
                        },
                    },
                    "errors": {
                        "$data": [],
                        "$type": {
                            "args": [{"args": [], "type": "str", "module": "builtins"}],
                            "type": "TList",
                            "module": "pytractions.base",
                        },
                    },
                    "o_out1": {
                        "$data": {"data": 0.0},
                        "$type": {
                            "args": [{"args": [], "type": "float", "module": "builtins"}],
                            "type": "Out",
                            "module": "pytractions.base",
                        },
                    },
                    "skip": False,
                    "skip_reason": "",
                    "state": "ready",
                    "stats": {
                        "$data": {"finished": "", "skipped": False, "started": ""},
                        "$type": {
                            "args": [],
                            "module": "pytractions.base",
                            "type": "TractionStats",
                        },
                    },
                    "uid": "1",
                },
                "$type": {
                    "args": [],
                    "module": "tests.test_base_traction",
                    "type": "test_tractor_to_json.<locals>.TTest1",
                },
            },
            "t_ttest2": {
                "$data": {
                    "a_reducer": {
                        "$data": {"a": 0.0},
                        "$type": {
                            "args": [{"args": [], "type": "float", "module": "builtins"}],
                            "type": "Arg",
                            "module": "pytractions.base",
                        },
                    },
                    "details": {
                        "$data": {},
                        "$type": {
                            "args": [
                                {"args": [], "type": "str", "module": "builtins"},
                                {"args": [], "type": "str", "module": "builtins"},
                            ],
                            "type": "TDict",
                            "module": "pytractions.base",
                        },
                    },
                    "errors": {
                        "$data": [],
                        "$type": {
                            "args": [{"args": [], "type": "str", "module": "builtins"}],
                            "type": "TList",
                            "module": "pytractions.base",
                        },
                    },
                    "i_in1": "TTest1[1]#o_out1",
                    "o_out1": {
                        "$data": {"data": 0.0},
                        "$type": {
                            "args": [{"args": [], "type": "float", "module": "builtins"}],
                            "type": "Out",
                            "module": "pytractions.base",
                        },
                    },
                    "skip": False,
                    "skip_reason": "",
                    "state": "ready",
                    "stats": {
                        "$data": {"finished": "", "skipped": False, "started": ""},
                        "$type": {
                            "args": [],
                            "module": "pytractions.base",
                            "type": "TractionStats",
                        },
                    },
                    "uid": "2",
                },
                "$type": {
                    "args": [],
                    "module": "tests.test_base_traction",
                    "type": "test_tractor_to_json.<locals>.TTest2",
                },
            },
            "t_ttest3": {
                "$data": {
                    "a_reducer": {
                        "$data": {"a": 0.0},
                        "$type": {
                            "args": [{"args": [], "type": "float", "module": "builtins"}],
                            "type": "Arg",
                            "module": "pytractions.base",
                        },
                    },
                    "details": {
                        "$data": {},
                        "$type": {
                            "args": [
                                {"args": [], "type": "str", "module": "builtins"},
                                {"args": [], "type": "str", "module": "builtins"},
                            ],
                            "type": "TDict",
                            "module": "pytractions.base",
                        },
                    },
                    "errors": {
                        "$data": [],
                        "$type": {
                            "args": [{"args": [], "type": "str", "module": "builtins"}],
                            "type": "TList",
                            "module": "pytractions.base",
                        },
                    },
                    "i_in1": "TTest2[2]#o_out1",
                    "o_out1": {
                        "$data": {"data": 0.0},
                        "$type": {
                            "args": [{"args": [], "type": "float", "module": "builtins"}],
                            "type": "Out",
                            "module": "pytractions.base",
                        },
                    },
                    "skip": False,
                    "skip_reason": "",
                    "state": "ready",
                    "stats": {
                        "$data": {"finished": "", "skipped": False, "started": ""},
                        "$type": {
                            "args": [],
                            "module": "pytractions.base",
                            "type": "TractionStats",
                        },
                    },
                    "uid": "3",
                },
                "$type": {
                    "args": [],
                    "module": "tests.test_base_traction",
                    "type": "test_tractor_to_json.<locals>.TTest2",
                },
            },
            "t_ttest4": {
                "$data": {
                    "a_reducer": {
                        "$data": {"a": 0.0},
                        "$type": {
                            "args": [{"args": [], "type": "float", "module": "builtins"}],
                            "type": "Arg",
                            "module": "pytractions.base",
                        },
                    },
                    "details": {
                        "$data": {},
                        "$type": {
                            "args": [
                                {"args": [], "type": "str", "module": "builtins"},
                                {"args": [], "type": "str", "module": "builtins"},
                            ],
                            "type": "TDict",
                            "module": "pytractions.base",
                        },
                    },
                    "errors": {
                        "$data": [],
                        "$type": {
                            "args": [{"args": [], "type": "str", "module": "builtins"}],
                            "type": "TList",
                            "module": "pytractions.base",
                        },
                    },
                    "i_in1": "TTest2[3]#o_out1",
                    "o_out1": {
                        "$data": {"data": 0.0},
                        "$type": {
                            "args": [{"args": [], "type": "float", "module": "builtins"}],
                            "type": "Out",
                            "module": "pytractions.base",
                        },
                    },
                    "skip": False,
                    "skip_reason": "",
                    "state": "ready",
                    "stats": {
                        "$data": {"finished": "", "skipped": False, "started": ""},
                        "$type": {
                            "args": [],
                            "module": "pytractions.base",
                            "type": "TractionStats",
                        },
                    },
                    "uid": "4",
                },
                "$type": {
                    "args": [],
                    "module": "tests.test_base_traction",
                    "type": "test_tractor_to_json.<locals>.TTest2",
                },
            },
            "tractions": {
                "$data": {
                    "t_ttest1": {
                        "$data": {
                            "a_multiplier": {
                                "$data": {"a": 10.0},
                                "$type": {
                                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                                    "module": "pytractions.base",
                                    "type": "Arg",
                                },
                            },
                            "details": {
                                "$data": {},
                                "$type": {
                                    "args": [
                                        {"args": [], "type": "str", "module": "builtins"},
                                        {"args": [], "type": "str", "module": "builtins"},
                                    ],
                                    "type": "TDict",
                                    "module": "pytractions.base",
                                },
                            },
                            "errors": {
                                "$data": [],
                                "$type": {
                                    "args": [{"args": [], "type": "str", "module": "builtins"}],
                                    "type": "TList",
                                    "module": "pytractions.base",
                                },
                            },
                            "o_out1": {
                                "$data": {"data": 10.0},
                                "$type": {
                                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                                    "type": "Out",
                                    "module": "pytractions.base",
                                },
                            },
                            "skip": False,
                            "skip_reason": "",
                            "state": "finished",
                            "stats": {
                                "$data": {
                                    "finished": "1990-01-01T00:00:01.00000Z",
                                    "skipped": False,
                                    "started": "1990-01-01T00:00:00.00000Z",
                                },
                                "$type": {
                                    "args": [],
                                    "module": "pytractions.base",
                                    "type": "TractionStats",
                                },
                            },
                            "uid": "tt1::1",
                        },
                        "$type": {
                            "args": [],
                            "module": "tests.test_base_traction",
                            "type": "test_tractor_to_json.<locals>.TTest1",
                        },
                    },
                    "t_ttest2": {
                        "$data": {
                            "a_reducer": {
                                "$data": {"a": 2.0},
                                "$type": {
                                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                                    "module": "pytractions.base",
                                    "type": "Arg",
                                },
                            },
                            "details": {
                                "$data": {},
                                "$type": {
                                    "args": [
                                        {"args": [], "type": "str", "module": "builtins"},
                                        {"args": [], "type": "str", "module": "builtins"},
                                    ],
                                    "type": "TDict",
                                    "module": "pytractions.base",
                                },
                            },
                            "errors": {
                                "$data": [],
                                "$type": {
                                    "args": [{"args": [], "type": "str", "module": "builtins"}],
                                    "type": "TList",
                                    "module": "pytractions.base",
                                },
                            },
                            "o_out1": {
                                "$data": {"data": 5.5},
                                "$type": {
                                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                                    "type": "Out",
                                    "module": "pytractions.base",
                                },
                            },
                            "i_in1": {
                                "$data": {"data": 10.0},
                                "$type": {
                                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                                    "type": "Out",
                                    "module": "pytractions.base",
                                },
                            },
                            "skip": False,
                            "skip_reason": "",
                            "state": "finished",
                            "stats": {
                                "$data": {
                                    "finished": "1990-01-01T00:00:03.00000Z",
                                    "skipped": False,
                                    "started": "1990-01-01T00:00:02.00000Z",
                                },
                                "$type": {
                                    "args": [],
                                    "module": "pytractions.base",
                                    "type": "TractionStats",
                                },
                            },
                            "uid": "tt1::2",
                        },
                        "$type": {
                            "args": [],
                            "module": "tests.test_base_traction",
                            "type": "test_tractor_to_json.<locals>.TTest2",
                        },
                    },
                    "t_ttest3": {
                        "$data": {
                            "a_reducer": {
                                "$data": {"a": 2.0},
                                "$type": {
                                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                                    "module": "pytractions.base",
                                    "type": "Arg",
                                },
                            },
                            "details": {
                                "$data": {},
                                "$type": {
                                    "args": [
                                        {"args": [], "type": "str", "module": "builtins"},
                                        {"args": [], "type": "str", "module": "builtins"},
                                    ],
                                    "type": "TDict",
                                    "module": "pytractions.base",
                                },
                            },
                            "errors": {
                                "$data": [],
                                "$type": {
                                    "args": [{"args": [], "type": "str", "module": "builtins"}],
                                    "type": "TList",
                                    "module": "pytractions.base",
                                },
                            },
                            "i_in1": {
                                "$data": {"data": 5.5},
                                "$type": {
                                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                                    "type": "Out",
                                    "module": "pytractions.base",
                                },
                            },
                            "o_out1": {
                                "$data": {"data": 3.25},
                                "$type": {
                                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                                    "type": "Out",
                                    "module": "pytractions.base",
                                },
                            },
                            "skip": False,
                            "skip_reason": "",
                            "state": "finished",
                            "stats": {
                                "$data": {
                                    "finished": "1990-01-01T00:00:05.00000Z",
                                    "skipped": False,
                                    "started": "1990-01-01T00:00:04.00000Z",
                                },
                                "$type": {
                                    "args": [],
                                    "module": "pytractions.base",
                                    "type": "TractionStats",
                                },
                            },
                            "uid": "tt1::3",
                        },
                        "$type": {
                            "args": [],
                            "module": "tests.test_base_traction",
                            "type": "test_tractor_to_json.<locals>.TTest2",
                        },
                    },
                    "t_ttest4": {
                        "$data": {
                            "a_reducer": {
                                "$data": {"a": 2.0},
                                "$type": {
                                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                                    "module": "pytractions.base",
                                    "type": "Arg",
                                },
                            },
                            "details": {
                                "$data": {},
                                "$type": {
                                    "args": [
                                        {"args": [], "type": "str", "module": "builtins"},
                                        {"args": [], "type": "str", "module": "builtins"},
                                    ],
                                    "type": "TDict",
                                    "module": "pytractions.base",
                                },
                            },
                            "errors": {
                                "$data": [],
                                "$type": {
                                    "args": [{"args": [], "type": "str", "module": "builtins"}],
                                    "type": "TList",
                                    "module": "pytractions.base",
                                },
                            },
                            "i_in1": {
                                "$data": {"data": 3.25},
                                "$type": {
                                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                                    "type": "Out",
                                    "module": "pytractions.base",
                                },
                            },
                            "o_out1": {
                                "$data": {"data": 2.125},
                                "$type": {
                                    "args": [{"args": [], "type": "float", "module": "builtins"}],
                                    "type": "Out",
                                    "module": "pytractions.base",
                                },
                            },
                            "skip": False,
                            "skip_reason": "",
                            "state": "finished",
                            "stats": {
                                "$data": {
                                    "finished": "1990-01-01T00:00:07.00000Z",
                                    "skipped": False,
                                    "started": "1990-01-01T00:00:06.00000Z",
                                },
                                "$type": {
                                    "args": [],
                                    "module": "pytractions.base",
                                    "type": "TractionStats",
                                },
                            },
                            "uid": "tt1::4",
                        },
                        "$type": {
                            "args": [],
                            "module": "tests.test_base_traction",
                            "type": "test_tractor_to_json.<locals>.TTest2",
                        },
                    },
                },
                "$type": {
                    "args": [
                        {"args": [], "type": "str", "module": "builtins"},
                        {"args": [], "type": "Traction", "module": "pytractions.base"},
                    ],
                    "module": "pytractions.base",
                    "type": "TDict",
                },
            },
            "uid": "tt1",
        },
        "$type": {
            "args": [],
            "module": "tests.test_base_traction",
            "type": "test_tractor_to_json.<locals>.TestTractor",
        },
    }


def test_type_from_to_json():
    original = TypeNode.from_type(Out[TList[Out[int]]])
    assert TypeNode.from_json(TypeNode.from_type(Out[TList[Out[int]]]).to_json()) == original


def test_type_in_out():
    assert TypeNode.from_type(Out[int]) == TypeNode.from_type(In[int])


def test_type_in_out_complex():
    assert TypeNode.from_type(Out[TList[Out[int]]]) == TypeNode.from_type(In[TList[In[int]]])


def test_type_in_out_complex_2():
    print(TypeNode.from_json(TypeNode.from_type(Out[TList[Out[int]]]).to_json()).to_json())
    assert TypeNode.from_json(
        TypeNode.from_type(Out[TList[Out[int]]]).to_json()
    ) == TypeNode.from_type(In[TList[In[int]]])



# -*- coding: utf-8 -*-

"""setup.py"""

from setuptools import setup, find_packages

from pytractions.pkgutils import traction_entry_points

import pytractions.transformations

classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
]

setup(
    name="pytractions",
    version="0.0.5",
    description="""Pytractions is python framework for modular programming.""",
    long_description="""Pytractions is python framework for modular programming.
    It is based on the concept of tractions, which are small, self-contained modules.
    Traction fundamental idea is code which is supposed to transform input data
    to output data while using external functionality provided by libraries, program or
    external services. Strong emphasis is put on typing and data validation.
    Indivudual tractions can be combined into pipelines, which can be executed
    locally, in contianer or as tekton task.
""",
    long_description_content_type="text/x-rst",
    author="Jindrich Luza",
    author_email="jluza@redhat.com",
    url="https://github.com/midnightercz/pytractions",
    classifiers=classifiers,
    packages=find_packages(exclude="tests"),
    data_files=[],
    install_requires=[
        "dill",
        "mypy",
        "dataclasses_json",
        "pyyaml",
        "lark",
        "jsonschema",
        "streamlit",
        "streamlit-extras",
        "streamlit_option_menu"
    ],
    include_package_data=True,
    entry_points={
        "tractions": [
            x for x in traction_entry_points(pytractions.transformations)
        ],
    }
)



import streamlit as st
from streamlit_extras.stylable_container import stylable_container

from pytractions.catalog import catalog

tractions, all_inputs, all_outputs, all_resources, all_args, tags = catalog()


st.title("Tractions catalog")

with st.popover("Filter"):

    name_filter = st.text_input("Name filter")

    tags_filter = st.multiselect("Tags filter", tags)

    inputs_filter = st.multiselect("Input types filter", all_inputs)

    outputs_filter = st.multiselect("Output types filter", all_outputs)

    resources_filter = st.multiselect("Resource types filter", all_resources)

    args_filter = st.multiselect("Arg types filter", all_args)

    type_filter = st.multiselect("Traction types filter", ["TRACTION", "TRACTOR", "STMD"])


def inputs_container():
    """Container for inputs."""
    return stylable_container(
        key="inputs_container",
        css_styles="""
{
    border: 1px solid rgba(49, 51, 63, 0.2);
    border-radius: 0.5rem;
    background-color: #c5dec8;
    padding: calc(1em - 1px);
}
""",
    )


def outputs_container():
    """Container for outputs."""
    return stylable_container(
        key="outputs_container",
        css_styles="""
{
    border: 1px solid rgba(49, 51, 63, 0.2);
    border-radius: 0.5rem;
    background-color: #dbd6c1;
    padding: calc(1em - 1px);
}
""",
    )


def resources_container():
    """Container for resources."""
    return stylable_container(
        key="resources_container",
        css_styles="""
{
    border: 1px solid rgba(49, 51, 63, 0.2);
    border-radius: 0.5rem;
    background-color: #dbc9c5;
    padding: calc(1em - 1px);
}
""",
    )


def args_container():
    """Container for args."""
    return stylable_container(
        key="args_container",
        css_styles="""
{
    border: 1px solid rgba(49, 51, 63, 0.2);
    border-radius: 0.5rem;
    background-color: #c1d0db;
    padding: calc(1em - 1px);
}
""",
    )


def indent_container():
    """Container with indent."""
    return stylable_container(
        key="indent_container",
        css_styles="""
{
    padding: calc(1em - 1px);
}
""",
    )


filtered_tractions, _, _, _, _, _ = catalog(
    tag_filter=tags_filter,
    type_filter=type_filter,
    name_filter=name_filter,
    inputs_type_filter=inputs_filter,
    outputs_type_filter=outputs_filter,
    resources_type_filter=resources_filter,
    args_type_filter=args_filter,
    type_to_str=True,
)

for d in filtered_tractions:
    for t in d["tractions"]:
        tc = stylable_container(
            key="tractions",
            css_styles="""
{
border: 1px solid rgba(49, 51, 63, 0.2);
padding: calc(1em - 1px);
border-radius: 0.5rem;
background-color: #f9f9f9;
}
        """,
        )
        with tc:
            tc.write(t["module"] + "." + t["name"])
            tc.write("description: " + (t["docs"] or ""))
            tc.multiselect(
                "Tags",
                key=d["name"] + "." + t["name"] + "-" + "tags",
                options=t["tags"],
                default=t["tags"],
                disabled=True,
            )
            tc.write("Args:")
            c = args_container()
            with c:
                for targ in t["args"]:
                    c_indent = indent_container()
                    c_indent.write(targ["name"] + " : " + targ["type"])
                    c_indent.write("description : " + (targ.get("docs") or ""))
            tc.write("Resources:")
            c = resources_container()
            with c:
                for tres in t["resources"]:
                    c_indent = indent_container()
                    c_indent.write(tres["name"] + " : " + tres["type"])
                    c_indent.write("description : " + (tres.get("docs") or ""))
            tc.write("Inputs:")
            c = inputs_container()
            with c:
                for tin in t["inputs"]:
                    c_indent = indent_container()
                    c_indent.write(tin["name"] + " : " + tin["type"])
                    c_indent.write("description : " + (tin.get("docs") or ""))
            tc.write("Outputs:")
            c = outputs_container()
            with c:
                for tout in t["outputs"]:
                    c_indent = indent_container()
                    c_indent.write(tin["name"] + " : " + tout["type"])
                    c_indent.write("description : " + (tout.get("docs") or ""))



import pytest

from pytractions.base import Traction, TIn, In, Out, OnUpdateCallable, Res, TRes, Base
from pytractions.tractor import Tractor


def test_tractor_attr():
    with pytest.raises(TypeError):

        class TT1(Tractor):
            i_in1: int

    with pytest.raises(TypeError):

        class TT2(Tractor):
            o_out1: int

    with pytest.raises(TypeError):

        class TT3(Tractor):
            a_arg1: int

    with pytest.raises(TypeError):

        class TT4(Tractor):
            r_res1: int

    with pytest.raises(TypeError):

        class TT5(Tractor):
            t_traction: int

    with pytest.raises(TypeError):

        class TT6(Tractor):
            custom_attribute: int


class Seq(Base):
    """Sequence resource."""

    val: int = 0

    def inc(self):
        """Return incremented value."""
        self.val += 1
        return self.val


class TestTraction(Traction):
    """Test Traction."""

    i_input: In[int]
    o_output: Out[int]
    r_seq: Res[Seq]

    def _run(self, on_update: OnUpdateCallable) -> None:
        self.o_output.data = self.i_input.data + self.r_seq.r.inc()


class TestTractor(Tractor):
    """Test Tractor."""

    i_in1: In[int] = TIn[int]()
    r_seq: Res[Seq] = TRes[Seq]()

    t_t1: TestTraction = TestTraction(uid="1", i_input=i_in1, r_seq=r_seq)

    o_out1: Out[int] = t_t1.o_output


class TestTractor2(Tractor):
    """Test Tractor2."""

    i_in1: In[int] = TIn()
    r_seq: Res[Seq] = TRes[Seq]()

    t_tractor1: TestTractor = TestTractor(uid="1", i_in1=i_in1, r_seq=r_seq)

    o_out1: Out[int] = t_tractor1.o_out1


def test_tractor_nested():
    seq = Seq(val=10)
    t = TestTractor2(uid="1", i_in1=In[int](data=1), r_seq=Res[Seq](r=seq))
    t.run()
    assert t.o_out1.data == 12
    assert t.tractions["t_tractor1"].o_out1.data == 12
    assert t.tractions["t_tractor1"].i_in1.data == 1



class TractionFailedError(Exception):
    """Exception indidating failure of a step."""


class UninitiatedResource(Exception):
    """Exception indidating resource is not initiated."""

    def __init__(self, msg):
        """Initialize the exception."""
        self.msg = msg



import argparse
import json
import logging
import sys
import os
import yaml

from .monitor import StructuredMonitor
from .base import Traction
from .runner_utils import parse_traction_str, gen_default_inputs


LOGGER = logging.getLogger()
sh = logging.StreamHandler(stream=sys.stdout)
sh.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
LOGGER.addHandler(sh)


class SimpleRunner:
    """Simple runner class."""

    def __init__(self, tractor, monitor_file):
        """Initialize the runner."""
        self.tractor = tractor
        self.monitor_file = monitor_file

    def run(self):
        """Run tractor."""
        monitor = StructuredMonitor(self.tractor, self.monitor_file)
        try:
            self.tractor.run(on_update=monitor.on_update)
        finally:
            monitor.close(self.tractor)

    def resubmit(self, traction):
        """Run tractor from specific traction."""
        loading_started = False
        _ttraction = getattr(self.tractor, traction)
        outputs = []

        for tf, ftype in self.tractor._fields.items():
            if tf == traction:
                loading_started = True
            if tf.startswith("t_") and loading_started:
                _ttraction = getattr(self.tractor, traction)
                for f, _ in _ttraction._fields.items():
                    if f.startswith("i_") and self.tractor._io_map[(traction, f)] not in outputs:
                        outputs.append(self.tractor._io_map[(traction, f)])
                traction_path = os.path.join(
                    self.monitor_file,
                    self.tractor.uid + "::" + getattr(self.tractor, tf).uid + ".json",
                )
                if os.path.exists(traction_path):
                    self.tractor.tractions[tf] = ftype.from_json(json.load(open(traction_path)))
        for output in outputs:
            if output[0] != "#":
                traction_path = os.path.join(
                    self.monitor_file,
                    self.tractor.uid + "::" + getattr(self.tractor, output[0]).uid + ".json",
                )
                self.tractor.tractions[output[0]] = getattr(self.tractor, output[0]).from_json(
                    json.load(open(traction_path))
                )
            else:
                traction_path = os.path.join(
                    self.monitor_file, self.tractor.uid + "::" + output[1] + ".json"
                )
                ftype = self.tractor._fields[output[1]]
                setattr(self.tractor, output[1], ftype.from_json(json.load(open(traction_path))))

        for f, ftype in self.tractor._fields.items():
            if f == traction:
                loading_started = True
            if f.startswith("t_") and loading_started:
                inputs = self.tractor._init_traction_input(f, ftype)
                for _in, t_in in inputs.items():
                    object.__setattr__(self.tractor.tractions[f], _in, t_in)

        monitor = StructuredMonitor(self.tractor, self.monitor_file)
        self.tractor.resubmit_from(traction)
        try:
            self.tractor.run(on_update=monitor.on_update)
        finally:
            monitor.close(self.tractor)


def run_main(args):
    """Run action."""
    traction_cls = parse_traction_str(args.traction)
    traction_init_fields = {}
    LOGGER.setLevel(getattr(logging, args.level))

    docs = yaml.safe_load_all(sys.stdin.read())
    for doc in docs:
        name, data, data_file = doc["name"], doc.get("data"), doc.get("data_file")
        if data_file:
            data = yaml.safe_load(open(data_file).read())
            data = data["data"]
        if name not in traction_cls._fields:
            raise AttributeError(f"{traction_cls.__name__} doesn't have field {name}")
        LOGGER.info(f"Loading input: {name}")
        traction_init_fields[name] = traction_cls._fields[name].content_from_json(
            yaml.safe_load(data)
        )
    traction = traction_cls(uid="0", **traction_init_fields)
    LOGGER.info("Running simple runner on directory {args.monitor}")
    runner = SimpleRunner(traction, args.monitor)
    runner.run()


def resubmit_main(args):
    """Resubmit action."""
    traction = Traction.from_json(json.load(open(os.path.join(args.monitor, "-root-.json"))))
    runner = SimpleRunner(traction, args.monitor)
    runner.resubmit(args.from_traction)


def gen_inputs_main(args):
    """Generate inputs for the traction."""
    traction_cls = parse_traction_str(args.traction)
    defaults = gen_default_inputs(traction_cls)
    yaml_out = []
    for k, v in defaults.items():
        yaml_out.append({"name": k, "data": v})
    print(yaml.dump_all(yaml_out))


def make_parsers(subparsers):
    """Make runner parser."""
    p_runner = subparsers.add_parser("local_run", help="Run pytraction module")
    p_runner.add_argument("traction", help="Path of traction to run (module:traction)", type=str)
    p_runner.add_argument(
        "--monitor", "-m", help="Path to monitor directory", type=str, default="monitor"
    )
    p_runner.add_argument(
        "--level",
        "-l",
        help="Set log level",
        type=str,
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
    )
    p_runner.set_defaults(command=run_main)

    p_resubmit = subparsers.add_parser("local_resubmit", help="Run pytraction module")
    p_resubmit.add_argument(
        "--monitor", "-m", help="Path to monitor directory", type=str, default="monitor"
    )
    p_resubmit.add_argument(
        "--from-traction", "-t", help="Resubmit from specific traction", type=str, default="monitor"
    )
    p_resubmit.set_defaults(command=resubmit_main)

    p_gen_inputs = subparsers.add_parser("generate_inputs", help="Run pytraction module")
    p_gen_inputs.add_argument(
        "traction", help="Path of traction to run (module:traction)", type=str
    )
    p_gen_inputs.set_defaults(command=gen_inputs_main)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Pytraction simple runner")
    subparsers = parser.add_subparsers(required=True, dest="command")
    make_parsers(subparsers)
    args = parser.parse_args()
    args.command(args)
